import warnings

from napari._qt.widgets._slider_compat import QDoubleSlider
from qtpy import QtGui
from qtpy.QtCore import Qt, QEvent
from scipy.ndimage.filters import gaussian_filter
from napari.layers import Image


class BlurSlider(QDoubleSlider):
    def __init__(self, viewer, image_layer=None, blur_func=None, parent=None):
        super().__init__(parent=parent)
        self.viewer = viewer
        self._smoothed_layer = None
        self.image_layer = image_layer
        self.setMaximum(20)
        self.valueChanged.connect(self.update_image)
        self.setMouseTracking(True)
        self.setOrientation(Qt.Horizontal)
        self.blur_func = blur_func if blur_func is not None else lambda img, val: gaussian_filter(img, val)
        self.children()[0].installEventFilter(self)

    def mousePressEvent(self, ev: QtGui.QMouseEvent) -> None:
        image_layer = self.image_layer
        if image_layer is None:
            return
        blurred = self.get_blurred_image()
        new_layer_name = "[smooth] %s" % image_layer.name
        if new_layer_name in self.viewer.layers:
            self._smoothed_layer = self.viewer.layers[new_layer_name]
            self._smoothed_layer.data = blurred
            self._smoothed_layer.translate = image_layer.translate[list(self.viewer.dims.displayed)]
            self._smoothed_layer.colormap = image_layer.colormap
            self._smoothed_layer.contrast_limits = image_layer.contrast_limits
            self._smoothed_layer.rgb = image_layer.rgb
        else:
            self._smoothed_layer = Image(
                blurred,
                name="[smooth] %s" % image_layer.name,
                translate=image_layer.translate[list(self.viewer.dims.displayed)],
                colormap=image_layer.colormap,
                contrast_limits=image_layer.contrast_limits,
                rgb=image_layer.rgb
            )
            self.viewer.add_layer(self._smoothed_layer)

    def mouseReleaseEvent(self, ev: QtGui.QMouseEvent) -> None:
        if self._smoothed_layer is None:
            return
        self.viewer.layers.remove(self._smoothed_layer)
        self._smoothed_layer = None

    def update_image(self, _):
        if self.image_layer is not None and self._smoothed_layer is not None:
            self._smoothed_layer.data = self.get_blurred_image()
            self._smoothed_layer.events.data()

    def get_blurred_image(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return self.blur_func(self.image_layer._data_view, self.value())

    def eventFilter(self, obj: 'QObject', event: 'QEvent') -> bool:
        if event.type() == QEvent.MouseButtonPress:
            self.mousePressEvent(event)
        elif event.type() == QEvent.MouseButtonRelease:
            self.mouseReleaseEvent(event)
        return super().eventFilter(obj, event)



import warnings

import numpy as np
from napari.utils.events import disconnect_events

from napari.utils.translations import trans
from qtpy.QtWidgets import QWidget
from qtpy.QtCore import Qt
from qtpy.QtGui import QPainter, QColor


class QtChangeableColorBox(QWidget):
    """A widget that shows a square with the current label color.

    Parameters
    ----------
    layer : napari.layers.Layer
        An instance of a napari layer.
    """

    def __init__(self, layer):
        super().__init__()

        self._layer = None
        self.layer = layer

        self.setAttribute(Qt.WA_DeleteOnClose)

        self._height = 24
        self.setFixedWidth(self._height)
        self.setFixedHeight(self._height)
        self.setToolTip(trans._('Selected label color'))

        self.color = None

    @property
    def layer(self):
        return self._layer

    @layer.setter
    def layer(self, new_layer):
        if new_layer == self.layer:
            return
        if self._layer is not None:
            self._layer.events.selected_label.disconnect(self._on_selected_label_change)
            self._layer.events.opacity.disconnect(self._on_opacity_change)
            self._layer.events.colormap.disconnect(self._on_colormap_change)
        self._layer = new_layer
        if new_layer is not None:
            new_layer.events.selected_label.connect(
                self._on_selected_label_change
            )
            new_layer.events.opacity.connect(self._on_opacity_change)
            new_layer.events.colormap.connect(self._on_colormap_change)

    def _on_selected_label_change(self, *args):
        """Receive layer model label selection change event & update colorbox."""
        self.update()

    def _on_opacity_change(self, *args):
        """Receive layer model label selection change event & update colorbox."""
        self.update()

    def _on_colormap_change(self, *args):
        """Receive label colormap change event & update colorbox."""
        self.update()

    def paintEvent(self, event):
        """Paint the colorbox.  If no color, display a checkerboard pattern.

        Parameters
        ----------
        event : qtpy.QtCore.QEvent
            Event from the Qt context.
        """
        painter = QPainter(self)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            selected_color = self.layer._selected_color if self.layer else None
        if selected_color is None:
            self.color = None
            for i in range(self._height // 4):
                for j in range(self._height // 4):
                    if (i % 2 == 0 and j % 2 == 0) or (
                        i % 2 == 1 and j % 2 == 1
                    ):
                        painter.setPen(QColor(230, 230, 230))
                        painter.setBrush(QColor(230, 230, 230))
                    else:
                        painter.setPen(QColor(25, 25, 25))
                        painter.setBrush(QColor(25, 25, 25))
                    painter.drawRect(i * 4, j * 4, 5, 5)
        else:
            color = np.multiply(selected_color, self.layer.opacity)
            color = np.round(255 * color).astype(int)
            painter.setPen(QColor(*list(color)))
            painter.setBrush(QColor(*list(color)))
            painter.drawRect(0, 0, self._height, self._height)
            self.color = tuple(color)

    def deleteLater(self):
        disconnect_events(self.layer.events, self)
        super().deleteLater()

    def closeEvent(self, event):
        """Disconnect events when widget is closing."""
        disconnect_events(self.layer.events, self)
        super().closeEvent(event)




__version__ = "0.1.0"

from ._widgets import AnnotatorWidget, InterpolationWidget
import napari
from packaging import version
if version.parse(napari.__version__) >= version.parse("0.4.15"):
    from ._widgets import ListWidgetBB
    __all__ = ["AnnotatorWidget", "InterpolationWidget", "ListWidgetBB"]
else:
    __all__ = ["AnnotatorWidget", "InterpolationWidget"]




import napari
import numpy as np
from scipy.ndimage import gaussian_gradient_magnitude
from skimage.filters import sobel, sobel_h, sobel_v
from scipy.signal import convolve2d
import cv2

from qtpy.QtCore import QRunnable, QThreadPool, Signal, QObject, Slot
import queue
import threading
import time


class FeatureExtractor:
    def __init__(self, max_threads=None):
        self.queueLock = threading.Lock()
        self.queue = queue.Queue()
        self.n_done = 0
        self.pool = QThreadPool.globalInstance()
        self.n_threads = self.pool.maxThreadCount() if max_threads is None else max_threads
        self.done_mask = None

    def start_jobs(self, img, outs, current_slice=None, dims_displayed=None, rgb=None, f=None):
        self.start = time.time()
        ndim = img.ndim - (1 if rgb else 0)
        current_slice = tuple(map(lambda s: 0 if type(s) == slice else s, current_slice))
        dims_not_displayed = tuple(i for i in range(ndim) if i not in dims_displayed)
        idx_list = np.indices((img.shape[i] if i in dims_not_displayed else 1 for i in range(ndim)))
        idx_list = idx_list.reshape((idx_list.shape[0], -1)).T
        if current_slice is not None:
            order = np.argsort(np.abs(idx_list-current_slice).sum(1))
            idx_list = idx_list[order]
        idx_list = list(map(lambda l: tuple(l[i] if i in dims_not_displayed else slice(None) for i in range(len(l))), idx_list))
        # viewer.dims.events.current_step.connect(on_current_step)
        self.init_runnables()
        self.done_mask = np.zeros([img.shape[i] for i in dims_not_displayed], bool)
        # Fill the queue
        self.queueLock.acquire()
        for idx in idx_list:
            self.queue.put(idx)
        self.queueLock.release()

        # Create new threads
        for runnable in self.runnables:
            runnable.data = img
            runnable.outs = outs
            runnable.rgb = rgb
            runnable.done_mask = self.done_mask
            runnable.dims_not_displayed = dims_not_displayed
            self.pool.start(runnable)

    def init_runnables(self):
        self.runnables = []
        self.n_done = 0
        for i in range(self.n_threads):
            runnable = self.FeatureExtractorTask(i, self.queue, self.queueLock)
            self.runnables.append(runnable)

    class FeatureExtractorTask(QRunnable):
        def __init__(self, threadID, queue, lock):
            super().__init__()
            self.threadID = threadID
            self.q = queue
            self.lock = lock
            self.data = None
            self.outs = None
            self.rgb = None
            self.done_mask = None
            self.dims_not_displayed = None
            self._signals = self.Signals()
            self.conv_filter_v = np.asarray([[0, -1, 1]]).astype(float)
            self.conv_filter_h = np.asarray([[0], [-1], [1]]).astype(float)

        @Slot()
        def run(self):
            if self.data is None or self.outs is None:
                return
            while True:
                self.lock.acquire()
                if not self.q.empty():
                    idx = self.q.get()
                    self.lock.release()
                    if not self.rgb:
                        # self.out[idx, ...] = gaussian_gradient_magnitude(self.data[idx].astype(float), 5)
                        self.outs[0][idx] = sobel_v(self.data[idx].astype(float))
                        self.outs[1][idx] = sobel_h(self.data[idx].astype(float))
                    else:
                        # grad_r = gaussian_gradient_magnitude(self.data[idx + (0,)].astype(float), 5)
                        # grad_g = gaussian_gradient_magnitude(self.data[idx + (1,)].astype(float), 5)
                        # grad_b = gaussian_gradient_magnitude(self.data[idx + (2,)].astype(float), 5)
                        r, g, b = self.data[idx + (0,)].astype(float), self.data[idx + (1,)].astype(float), self.data[idx + (2,)].astype(float)
                        channels_v = []
                        channels_h = []
                        if np.any(r):
                            channels_v.append(cv2.filter2D(r, -1, self.conv_filter_v).astype(float))
                            channels_h.append(cv2.filter2D(r, -1, self.conv_filter_h))
                        if np.any(g):
                            channels_v.append(cv2.filter2D(g, -1, self.conv_filter_v))
                            channels_h.append(cv2.filter2D(g, -1, self.conv_filter_h))
                        if np.any(b):
                            channels_v.append(cv2.filter2D(b, -1, self.conv_filter_v))
                            channels_h.append(cv2.filter2D(b, -1, self.conv_filter_h))
                        self.outs[0][idx] = sum(channels_v)
                        self.outs[1][idx] = sum(channels_h)

                    self.q.task_done()
                    self.slice_done.emit(idx)
                    self.done_mask[tuple(idx[i] for i in self.dims_not_displayed)] = True
                else:
                    self.lock.release()
                    break
            self.done.emit()

        @property
        def done(self):
            return self._signals.done

        @property
        def slice_done(self):
            return self._signals.slice_done

        class Signals(QObject):
            done = Signal()
            slice_done = Signal(object)



import sys

import napari
from napari.utils.notifications import notification_manager
from qtpy.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QAction, QPlainTextEdit
from qtpy.QtCore import QRegularExpression, Qt, Signal, QObject, QThread
from qtpy.QtGui import QTextCharFormat, QFont, QSyntaxHighlighter, QColor
from qtpy import QtCore, QtGui
import numpy as np
import skimage
import warnings
import keyword

from ._utils import ProgressWidget


class ScriptWorker(QObject):
    done = Signal("PyQt_PyObject")
    script = None
    image = None

    def run(self):
        if self.script is None:
            raise ValueError("No script was set!")
        if self.image is None:
            raise ValueError("'image' was None!")
        locals = {"image": self.image}
        try:
            exec(self.script, {"np": np, "skimage": skimage}, locals)
            if "features" in locals:
                features = locals["features"]
                if features is not None and features.ndim == 2:
                    features = np.concatenate([features[..., np.newaxis]] * 3, -1)
                self.done.emit(features)
            else:
                warnings.warn("The output should be stored in a variable called 'features'")
                self.done.emit(None)
        except Exception as e:
            self.done.emit(None)
            raise e



class PythonHighlighter(QSyntaxHighlighter):
    class HighlightingRule:
        pattern = QRegularExpression()
        format = QTextCharFormat()
    highlightingRules = []
    commentFormat = QTextCharFormat()
    keywordFormat = QTextCharFormat()
    classFormat = QTextCharFormat()
    quotationFormat = QTextCharFormat()
    singleLineCommentFormat = QTextCharFormat()
    tripleQuotationFormat = QTextCharFormat()
    functionFormat = QTextCharFormat()
    numericFormat = QTextCharFormat()
    
    def __init__(self, parent):
        super().__init__(parent)
        self.keywordFormat.setForeground(QColor("darkorange"))
        self.keywordFormat.setFontWeight(QFont.Bold)
        keywordPatterns = ["\\b%s\\b" % kw for kw in keyword.kwlist]
        for pattern in keywordPatterns:
            rule = self.HighlightingRule()
            rule.pattern = QRegularExpression(pattern)
            rule.format = self.keywordFormat
            self.highlightingRules.append(rule)

        rule = self.HighlightingRule()
        self.numericFormat.setForeground(QColor("deepskyblue"))
        rule.pattern = QRegularExpression("(0((b[01]+)|(O[0-7]+)|(x[0-9A-Fa-f]+)))|([0-9]*\\.?[0-9]*j?)")
        rule.format = self.numericFormat
        self.highlightingRules.append(rule)

        rule = self.HighlightingRule()
        self.functionFormat.setForeground(QColor("gold"))
        rule.pattern = QRegularExpression("\\b[a-z][A-Za-z0-9_]*(?=\\()")
        rule.format = self.functionFormat
        self.highlightingRules.append(rule)

        rule = self.HighlightingRule()
        self.classFormat.setForeground(QColor("lightblue"))
        rule.pattern = QRegularExpression("\\b[A-Z][A-Za-z0-9_]*(?=\\()")
        rule.format = self.classFormat
        self.highlightingRules.append(rule)

        rule = self.HighlightingRule()
        self.quotationFormat.setForeground(Qt.darkGreen)
        rule.pattern = QRegularExpression("([\"'])(?:(?=(\\\\?))\\2.)*?\\1")
        rule.format = self.quotationFormat
        self.highlightingRules.append(rule)

        rule = self.HighlightingRule()
        self.commentFormat.setForeground(Qt.gray)
        rule.pattern = QRegularExpression("#.*")
        rule.format = self.commentFormat
        self.highlightingRules.append(rule)

    def highlightBlock(self, text):
        for rule in self.highlightingRules:
            matchIterator = rule.pattern.globalMatch(text)
            while matchIterator.hasNext():
                match = matchIterator.next()
                self.setFormat(match.capturedStart(), match.capturedLength(), rule.format)


class CodeEditor(QPlainTextEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setStyleSheet("font-family:'Courier New'; background-color: black")

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        if event.key() == QtCore.Qt.Key_Tab:
            self.insertPlainText("    ")
        else:
            super().keyPressEvent(event)


class ImageProcessingWidget(QWidget):
    try_script = Signal()

    def __init__(self, image, viewer: napari.Viewer):
        super().__init__()
        self.viewer = viewer
        self.progress_dialog = ProgressWidget(message="Calculating feature, please wait...")
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            viewer.window.qt_viewer.window().destroyed.connect(self.on_destroy)
        self.script_thread = QThread()
        self.script_worker = ScriptWorker()
        self.script_worker.moveToThread(self.script_thread)
        self.script_worker.done.connect(self.script_thread.quit)
        self.script_worker.done.connect(lambda _: self.progress_dialog.setVisible(False))
        self.script_worker.done.connect(lambda: self.run_button.setEnabled(True))
        self.script_worker.done.connect(lambda: self.try_button.setEnabled(True))
        self.script_thread.started.connect(self.script_worker.run)
        self.script_thread.finished.connect(lambda: self.progress_dialog.setVisible(False))
        layout = QVBoxLayout()
        self.text_settings = QtCore.QSettings()
        self.text_edit = CodeEditor()
        self.text_edit.document().setPlainText(self.text_settings.value("img_proc_script", ""))
        self.highlighter = PythonHighlighter(self.text_edit.document())
        font_size = self.text_settings.value("script_font_size")
        if font_size:
            font = self.text_edit.font()
            font.setPointSize(font_size)
            self.text_edit.setFont(font)
        run_action_enter = QAction(self.text_edit)
        run_action_enter.setAutoRepeat(False)
        run_action_enter.setShortcut("Ctrl+Enter")
        run_action_enter.triggered.connect(self.execute)
        self.text_edit.addAction(run_action_enter)

        run_action_return = QAction(self.text_edit)
        run_action_return.setAutoRepeat(False)
        run_action_return.setShortcut("Ctrl+Return")
        run_action_return.triggered.connect(self.execute)
        self.text_edit.addAction(run_action_return)

        increase_fontsize_action = QAction(self.text_edit)
        increase_fontsize_action.setShortcut(QtGui.QKeySequence(QtCore.Qt.CTRL+QtCore.Qt.Key_Plus))
        increase_fontsize_action.triggered.connect(self.increase_font_size)
        self.text_edit.addAction(increase_fontsize_action)

        decrease_fontsize_action = QAction(self.text_edit)
        decrease_fontsize_action.setShortcut(QtGui.QKeySequence(QtCore.Qt.CTRL+QtCore.Qt.Key_Minus))
        decrease_fontsize_action.triggered.connect(self.decrease_font_size)
        self.text_edit.addAction(decrease_fontsize_action)
        layout.addWidget(self.text_edit)

        buttons_layout = QHBoxLayout()
        self.run_button = QPushButton("Set")
        self.run_button.clicked.connect(self.execute)
        buttons_layout.addWidget(self.run_button)
        self.try_button = QPushButton("Try")
        self.try_button.clicked.connect(self.try_code)
        buttons_layout.addWidget(self.try_button)
        layout.addLayout(buttons_layout)
        self.setLayout(layout)
        self.image = image
        self.features = image

    def run_script(self):
        self.run_button.setEnabled(False)
        self.try_button.setEnabled(False)
        script = self.text_edit.document().toPlainText()
        self.text_settings.setValue("img_proc_script", script)
        if self.image is None:
            warnings.warn("image is None")
            return None
        self.script_worker.script = script
        self.script_worker.image = self.image.copy() if self.image is not None else None
        self.progress_dialog.setVisible(True)
        self.script_thread.start()

    def set_features(self, features):
        if features is None:
            return
        self.features = features.astype(float)
        self.script_worker.done.disconnect(self.set_features)

    def execute(self):
        self.script_worker.done.connect(self.set_features)
        self.run_script()

    def display_features(self, features):
        if features is not None:
            if "Feature map" not in self.viewer.layers:
                self.viewer.add_image(
                    features,
                    name="Feature map",
                    rgb=True
                )
            else:
                self.viewer.layers["Feature map"].data = features
        self.script_worker.done.disconnect(self.display_features)

    def try_code(self):
        self.try_script.emit()
        self.script_worker.done.connect(self.display_features)
        self.run_script()

    def increase_font_size(self):
        font = self.text_edit.font()
        font.setPointSize(font.pointSize()+1)
        self.text_edit.setFont(font)
        self.text_settings.setValue("script_font_size", font.pointSize())

    def decrease_font_size(self):
        font = self.text_edit.font()
        font.setPointSize(font.pointSize()-1)
        self.text_edit.setFont(font)
        self.text_settings.setValue("script_font_size", font.pointSize())

    def on_destroy(self):
        try:
            self.progress_dialog.close()
        except:
            ...



try:
    import MinArea
except ImportError:
    MinArea = None


if MinArea is not None:
    from time import sleep

    import tifffile
    from napari.utils.notifications import show_info

    from ._utils import WidgetWithLayerList, CollapsibleWidget, CollapsibleWidgetGroup, ProgressWidget, QDoubleSlider
    from ._utils.delayed_executor import DelayedExecutor
    from ._utils.callbacks import keep_layer_on_top
    import napari
    from napari.layers import Image, Labels, Points, Shapes
    from napari.utils.colormaps.standardize_color import transform_color
    # from napari._qt.widgets._slider_compat import QDoubleSlider
    from qtpy.QtCore import Signal, Qt, QObject, QThread
    from qtpy.QtWidgets import (
        QVBoxLayout,
        QHBoxLayout,
        QLabel,
        QDoubleSpinBox,
        QCheckBox,
        QPushButton,
        QFrame,
        QWidget,
        QApplication,
        QComboBox,
        QSpinBox,
        QSizePolicy
    )
    from threading import Thread
    import cv2
    import numpy as np
    import skimage.measure, skimage.morphology
    from scipy.ndimage import gaussian_filter, zoom
    import itertools
    import colorsys
    from random import shuffle, seed
    import SimpleITK as sitk


    def pts_2_bb(p1, p2, image_size, scale=1.):
        center = (p1 + p2) / 2
        size = np.sqrt(((p1 - p2) ** 2).sum()) + 10
        size = np.divide(size, scale)
        bb = np.asarray(
            np.where(list(itertools.product((False, True), repeat=3)), center + size / 2, center - size / 2))
        bb = np.clip(bb, 0, np.asarray(image_size) - 1)
        return bb


    def get_bb_corners(bb):
        return np.concatenate([bb.min(0, keepdims=True), bb.max(0, keepdims=True)])


    def bb_2_slice(bb):
        bb_corners = get_bb_corners(bb).round().astype(int)
        return tuple(slice(bb_corners[0, i], bb_corners[1, i]) for i in range(3))


    def color_to_hex_string(rgb_tuple):
        return '#%.2X%.2X%.2X' % rgb_tuple


    def generate_label_colors(n):
        rgb_tuples = [colorsys.hsv_to_rgb(x * 1.0 / n, 1., 1.) for x in range(n)]
        rgb_tuples = [(int(rgb_tuples[i][0] * 255), int(rgb_tuples[i][1] * 255), int(rgb_tuples[i][2] * 255)) for i in
                      range(n)]
        seed(0)
        shuffle(rgb_tuples)
        return rgb_tuples


    def generate_sphere_mask(shape):
        ndim = len(shape)
        r = max(shape) // 2
        mask = skimage.morphology.ball(r)
        shape_diffs = tuple(shape[i] - mask.shape[i] for i in range(ndim))
        crop_idx = tuple(
            slice(max(0, round(-shape_diffs[i] / 2 + 0.1)), mask.shape[i] + min(0, round(shape_diffs[i] / 2 + 0.1))) for
            i in range(ndim))
        mask = mask[crop_idx]
        mask = np.pad(mask, [(max(0, round(shape_diffs[i] / 2)), -min(0, round(-shape_diffs[i] / 2 + 0.1))) for i in
                             range(ndim)])
        assert mask.shape == shape
        return mask


    class ColorPairsCallback:
        def __init__(self, n_colors=50):
            self.prev_selection = set()
            self.prev_len = 0
            self.prev_data = None
            color_cycle = list(
                map(lambda color_tuple: color_to_hex_string(color_tuple), generate_label_colors(n_colors)))
            self.color_cycle = itertools.cycle(color_cycle)

        def __call__(self, event):
            points_layer = event.source
            if len(points_layer.selected_data) > 1:
                points_layer.selected_data = self.prev_selection
                points_layer.refresh()
                return
            self.prev_selection = points_layer.selected_data.copy()
            if event.type == "data":
                if len(points_layer.data) > self.prev_len:
                    if len(points_layer.data) % 2 == 1:
                        points_layer.current_face_color = next(self.color_cycle)
                        points_layer.face_color[-1] = transform_color(points_layer.current_face_color)
                    self.prev_data = points_layer.data.copy()
                else:
                    removed_idx = np.squeeze(np.argwhere(
                        np.all(np.any(~np.equal(self.prev_data[np.newaxis], points_layer.data[:, np.newaxis]), -1),
                               axis=0)))
                    if removed_idx.ndim == 0:
                        removed_idx = int(removed_idx)
                        points_layer.selected_data.clear()
                        if removed_idx % 2 == 0 and removed_idx < len(points_layer.data):
                            points_layer.selected_data.add(removed_idx)
                        elif removed_idx % 2 == 1:
                            points_layer.selected_data.add(removed_idx - 1)
                        points_layer.remove_selected()
                points_layer.refresh()
                self.prev_data = points_layer.data
                self.prev_len = len(points_layer.data)


    it = [-1]
    class EstimationWorker(QObject):
        image_data_received = Signal(str, "PyQt_PyObject", "PyQt_PyObject")
        mask_data_received = Signal("PyQt_PyObject", "PyQt_PyObject")
        remove_layer = Signal(str)
        layer_invalidated = Signal(str)
        all_done = Signal()
        annotation_needed_signal = Signal("PyQt_PyObject", "PyQt_PyObject")
        MANUAL_ANNOTATION = "Manual"
        THRESHOLD_ANNOTATION = "Threshold"
        MINIMAL_CONTOUR_ANNOTATION = "Minimal contour"
        ANNOTATION_METHODS = [MANUAL_ANNOTATION, THRESHOLD_ANNOTATION, MINIMAL_CONTOUR_ANNOTATION]

        def __init__(self, viewer, minimal_contour_widget, parent=None):
            super().__init__(parent)
            self.viewer = viewer
            self.minimal_contour_widget = minimal_contour_widget
            self.center_mat = None
            self.rotation_matrix = None
            self.translation_matrix = None
            self.image = None
            self.points = None
            self.use_correction = None
            self.beta = None
            self.alpha = None
            self.done_pressed = False
            self.blur_func = None
            self.init_slice_annotation_fun = None
            self.slice_image_layer = None
            self.slice_labels_layer = None
            self.slice_annotation = None
            self.manual_annotation_done = False
            self.z_scale = None
            #create annotation dialogs
            self.manual_annotation_dialog = QWidget()
            layout = QVBoxLayout()
            layout.addWidget(QLabel("Annotate image, then click 'Done'"))
            annotation_done_button = QPushButton("Done")
            layout.addWidget(annotation_done_button)
            def on_done_pressed():
                self.done_pressed = True
            annotation_done_button.clicked.connect(self.finish_manual_annotation)
            annotation_done_button.clicked.connect(on_done_pressed)
            self.manual_annotation_dialog.setLayout(layout)
            self.manual_annotation_dialog.setWindowFlags(Qt.WindowStaysOnTopHint | Qt.Dialog)
            self.manual_annotation_dialog.setAttribute(Qt.WA_ShowWithoutActivating)

            self.threshold_annotation_dialog = QWidget()
            layout = QVBoxLayout()
            layout.addWidget(QLabel("Threshold:"))
            threshold_slider = QDoubleSlider(Qt.Horizontal)
            threshold_slider.valueChanged.connect(self.threshold_value_changed)
            layout.addWidget(threshold_slider)
            annotation_done_button = QPushButton("Done")
            layout.addWidget(annotation_done_button)
            annotation_done_button.clicked.connect(self.finish_threshold_annotation)
            annotation_done_button.clicked.connect(on_done_pressed)
            self.threshold_annotation_dialog.setLayout(layout)
            self.annotation_needed_signal.connect(self.begin_slice_annotation)
            self.slice_annotation_methods = {EstimationWorker.MANUAL_ANNOTATION: self.manual_annotation, EstimationWorker.THRESHOLD_ANNOTATION: self.threshold_annotation, EstimationWorker.MINIMAL_CONTOUR_ANNOTATION: self.minimal_contour_annotation}
        
        def begin_slice_annotation(self, image_slice, distance_map_slice):
            self.slice_image_layer = self.viewer.add_image(image_slice)
            self.slice_labels_layer = self.viewer.add_labels(np.zeros_like(image_slice, dtype=int))
            self.slice_labels_layer.brush_size = 1
            self.slice_labels_layer.mode = "paint"
            self.init_slice_annotation_fun()
        
        def finish_slice_annotation(self):
            self.slice_annotation = self.slice_labels_layer.data > 0
            self.viewer.layers.remove(self.slice_image_layer)
            self.viewer.layers.remove(self.slice_labels_layer)
            self.slice_labels_layer = None
            self.slice_image_layer = None

        def manual_annotation(self):
            self.manual_annotation_dialog.show()
        
        def finish_manual_annotation(self):
            self.manual_annotation_dialog.hide()
            self.finish_slice_annotation()
        
        def threshold_annotation(self):
            threshold_slider = self.threshold_annotation_dialog.children()[2]
            threshold_slider.setMinimum(self.slice_image_layer.data.min())
            threshold_slider.setMaximum(self.slice_image_layer.data.max())
            threshold_slider.setValue((threshold_slider.minimum()+threshold_slider.maximum())/2)
            self.threshold_annotation_dialog.show()

        def threshold_value_changed(self, val):
            self.slice_labels_layer.data = self.slice_image_layer.data >= val

        def finish_threshold_annotation(self):
            self.threshold_annotation_dialog.hide()
            self.finish_slice_annotation()
        
        def minimal_contour_annotation(self):
            self.viewer.layers.selection.select_only(self.minimal_contour_widget.anchor_points)
            self.minimal_contour_widget.image.layer = self.slice_image_layer
            self.minimal_contour_widget.labels.layer = self.slice_labels_layer
            self.manual_annotation_dialog.show()
        
        def finish_minimal_contour_annotation(self):
            self.manual_annotation_dialog.hide()
            self.finish_slice_annotation()

        def initialize(self, image, points, use_correction, beta, alpha, slice_annotation_method="Manual", blur_func=None, z_scale=1.):
            self.image = image
            self.points = points
            self.use_correction = use_correction
            self.beta = beta
            self.alpha = alpha
            self.blur_func = blur_func
            self.init_slice_annotation_fun = self.slice_annotation_methods[slice_annotation_method]
            self.z_scale = z_scale

        def run(self):
            import time
            start = time.time()
            print("started")
            postscript = ""
            estimator = MinArea.MinimalSurfaceCalculator()
            # estimator.hook_stage_data_init_event(
            #     MinArea.AREA_EIKONAL_STAGE,
            #     lambda arr, idx: self.data_initializer(
            #         "Area Eikonal %d%s" % (0, postscript),
            #         0,
            #         {
            #             'colormap': "plasma",
            #             'translate': offset,
            #             "visible": False
            #         }
            #     )(arr, idx)
            # )
            # estimator.hook_stage_iteration_event(MinArea.AREA_EIKONAL_STAGE,
            #                                      lambda idx: self.data_updater("Area Eikonal %d%s" % (0, postscript))(
            #                                          idx))
            # estimator.hook_stage_finished_event(MinArea.AREA_EIKONAL_STAGE,
            #                                     lambda: self.data_finalizer("Area Eikonal %d%s" % (0, postscript))())
            # estimator.hook_stage_data_init_event(
            #     MinArea.AREA_EIKONAL_STAGE,
            #     lambda arr, idx: self.data_initializer(
            #         "Area Eikonal %d%s" % (1, postscript),
            #         1,
            #         {
            #             'colormap': "plasma",
            #             'translate': offset,
            #             "visible": False
            #         }
            #     )(arr, idx)
            # )
            # estimator.hook_stage_iteration_event(MinArea.AREA_EIKONAL_STAGE,
            #                                      lambda idx: self.data_updater("Area Eikonal %d%s" % (1, postscript))(
            #                                          idx))
            # estimator.hook_stage_finished_event(MinArea.AREA_EIKONAL_STAGE,
            #                                     lambda: self.data_finalizer("Area Eikonal %d%s" % (1, postscript))())
            # self.hook_callbacks(estimator, MinArea.AREA_EIKONAL_STAGE, "Area Eikonal 0", {'colormap': "plasma", 'translate': offset, "visible": False}, 0)
            # self.hook_callbacks(estimator, MinArea.AREA_EIKONAL_STAGE, "Area Eikonal 1", {'colormap': "plasma", 'translate': offset, "visible": False}, 1)
            # self.hook_callbacks(estimator, MinArea.ROTATED_AREA_EIKONAL_STAGE, "Rotated Area Eikonal", {'colormap': "plasma"}, 0)

            # estimator.hook_stage_data_init_event(
            #     MinArea.PLANE_PHASEFIELD_STAGE,
            #     lambda arr, idx: self.data_initializer("Plane PhaseField%s" % postscript,
            #                                            layer_args={
            #                                                'colormap': "plasma",
            #                                                'translate': offset + self.translation_matrix +
            #                                                             np.asarray([0, 0, self.center_mat[
            #                                                                 -1]]) @ self.rotation_matrix.T,
            #                                                'rotate': self.rotation_matrix,
            #                                                'opacity': 0.6,
            #                                                "visible": False
            #                                            }
            #                                            )(arr, idx)
            # )
            # estimator.hook_stage_iteration_event(MinArea.PLANE_PHASEFIELD_STAGE,
            #                                      lambda idx: self.data_updater("Plane PhaseField%s" % postscript)(idx))
            # estimator.hook_stage_finished_event(MinArea.PLANE_PHASEFIELD_STAGE,
            #                                     lambda: self.data_finalizer("Plane PhaseField%s" % postscript)())

            # estimator.hook_stage_data_init_event(
            #     MinArea.TRANSPORT_FUNCTION_STAGE,
            #     lambda arr, idx: self.data_initializer("Transport Function%s" % postscript,
            #                                            0,
            #                                            {
            #                                                'colormap': "turbo",
            #                                                'translate': offset + self.translation_matrix,
            #                                                'rotate': self.rotation_matrix,
            #                                                'opacity': 0.5,
            #                                                'rendering': "iso",
            #                                                "iso_threshold": 0.
            #                                            })(arr, idx)
            # )
            # # estimator.hook_stage_data_init_event(MinArea.TRANSPORT_FUNCTION_STAGE, lambda arr, idx: print("tf init"))
            # estimator.hook_stage_iteration_event(MinArea.TRANSPORT_FUNCTION_STAGE,
            #                                      lambda idx: self.data_updater("Transport Function%s" % postscript)(
            #                                          idx))
            # # estimator.hook_stage_iteration_event(MinArea.TRANSPORT_FUNCTION_STAGE, lambda idx: print("tf iter %d" % idx))
            # estimator.hook_stage_finished_event(MinArea.TRANSPORT_FUNCTION_STAGE,
            #                                     lambda: self.data_finalizer("Transport Function%s" % postscript)())
            #
            def tform_calculated(rotation, translation):
                self.rotation_matrix = np.flip(rotation.copy()).reshape(3, 3)
                self.translation_matrix = np.flip(translation.copy())

            estimator.hook_transform_calculated_event(tform_calculated)

            # def center_calculated(center):
            #     self.center_mat = np.flip(center.copy())
            #
            # estimator.hook_plane_center_calculated_event(center_calculated)
            estimator.set_initial_plane_calculator(self.segment_initial_slice)
            print("callbacks attached in %d seconds" % (time.time() - start))
            timestamp = time.time()
            for i in range(len(self.points) // 2):
                bounding_box = pts_2_bb(self.points[2 * i], self.points[2 * i + 1], self.image.shape)
                bounding_box = np.clip(bounding_box, 0, np.asarray(self.image.shape) - 1).round().astype(int)
                offset = bounding_box.min(0, keepdims=True)
                bb_slice = bb_2_slice(bounding_box)
                point1 = np.flip(self.points[2 * i] - offset).reshape(-1)
                point2 = np.flip(self.points[2 * i + 1] - offset).reshape(-1)
                point1[2] *= self.z_scale
                point2[2] *= self.z_scale
                print(point1, point2)
                data = self.image[bb_slice]
                data = zoom(data, (self.z_scale, 1, 1))
                data = (data - data.min()) / (data.max() - data.min())
                # data = scipy.ndimage.gaussian_filter(data, self.blur_sigma)
                if self.blur_func is not None:
                    data = self.blur_func(data)
                offset = np.clip(bounding_box.min(0), 0, np.asarray(self.image.shape) - 1)
                # start = time.time()

                output = estimator.calculate(data, point1, point2, self.use_correction, self.beta, self.alpha, 1000)
                tifffile.imwrite("C:\\Users\\Tejfel\\Downloads\\tf.tiff", output)
                tifffile.imwrite("C:\\Users\\Tejfel\\Downloads\\zoomed.tiff", data)
                segmented = (output >= 0)
                labelled = skimage.measure.label(segmented)
                obj_pixel = np.argwhere(output == output.max())[0]
                obj_label = labelled[tuple(obj_pixel)]
                mask = (labelled == obj_label) * (i + 1)
                mask = zoom(mask, (1/self.z_scale, 1, 1), order=0)
                # self.image_data_received.emit("Result", output, {"colormap": "plasma", "translate": offset})
                self.mask_data_received.emit(mask, offset)
                postscript = " - %d" % (i + 1)
                # self.remove_layer.emit("Transport Function")
            show_info("Estimation done in %d" % (time.time()-start))
            self.all_done.emit()

        def data_initializer(self, name, selected_idx=None, layer_args=None):
            if layer_args is None:
                layer_args = {}

            def initialize(arr, idx):
                try:
                    if selected_idx in [None, idx]:
                        self.image_data_received.emit(name, arr, layer_args)
                    else:
                        print("%s was not initialized as id (%d) was not %d" % (name, idx, selected_idx))
                except Exception as e:
                    ...

            return initialize
            # return lambda arr, idx: print("initializing", name)

        def data_updater(self, name):
            def update_viewer(iteration):
                it[0] += 1
                if it[0] % 100 == 0: #TODO undo
                    self.layer_invalidated.emit(name)
            return update_viewer

        def data_finalizer(self, name):
            def finalize():
                if name not in self.viewer.layers:
                    print("%s not in layer list" % name)
                    return
                self.viewer.layers[name].data = np.copy(self.viewer.layers[name].data)

            return finalize

        def segment_initial_slice(self, image_slice, distance_map):

            image_slice = np.squeeze(image_slice)
            distance_map = np.squeeze(distance_map)
            self.annotation_needed_signal.emit(image_slice, distance_map)
            while not self.done_pressed:
                sleep(0.1)
                QApplication.processEvents()
            self.done_pressed = False
            mask = self.slice_annotation.astype(float)
            self.slice_annotation = None
            return mask

    class MinimalSurfaceWidget(WidgetWithLayerList):
        def __init__(self, viewer: napari.Viewer, minimal_contour_widget):
            super().__init__(viewer, [("image", Image), ("labels", Labels)])
            self.blurred_layer = None
            self.progress_widget = ProgressWidget(message="blurring...")
            self.minimal_contour_widget = minimal_contour_widget
            self.points_layer = self.viewer.add_points(ndim=3, size=2, name="3D Anchors [DO NOT REMOVE]")
            self.points_layer.events.connect(ColorPairsCallback())
            self.points_layer.mouse_drag_callbacks.append(self.drag_points_callback)
            self.viewer.layers.events.connect(keep_layer_on_top(self.points_layer))
            layout = QVBoxLayout()
            layout.setAlignment(Qt.AlignTop)
            # Parameters for the estimator
            params_widget = CollapsibleWidget("Image features")
            params_widget.setContentsMargins(0, 0, 0, 0)
            params_layout = QVBoxLayout()
            self.image_feature_combobox = QComboBox(self)
            self.image_feature_combobox.addItem("Gradient")
            params_layout.addWidget(self.image_feature_combobox)

            alpha_layout = QHBoxLayout()
            alpha_layout.addWidget(QLabel("\u03B1", parent=self))
            self.alpha_slider = QDoubleSlider(parent=self)
            self.alpha_slider.setMinimum(0.)
            self.alpha_slider.setMaximum(1.)
            self.alpha_slider.setValue(0.01)
            self.alpha_slider.setDecimals(3)
            self.alpha_slider.setOrientation(Qt.Horizontal)
            alpha_layout.addWidget(self.alpha_slider)
            params_layout.addLayout(alpha_layout)

            beta_layout = QHBoxLayout()
            beta_layout.addWidget(QLabel("\u03B2", parent=self))
            self.beta_spinner = QDoubleSpinBox()
            self.beta_spinner.setMinimum(1.)
            self.beta_spinner.setMaximum(100.)
            self.beta_spinner.setValue(30.)
            self.beta_spinner.setSingleStep(1.)
            self.beta_spinner.setSizePolicy(QSizePolicy.Policy.Expanding, self.beta_spinner.sizePolicy().verticalPolicy())
            beta_layout.addWidget(self.beta_spinner)
            params_layout.addLayout(beta_layout)

            z_scale_layout = QHBoxLayout()
            z_scale_layout.addWidget(QLabel("Z-scale"))
            self.z_scale_spinbox = QDoubleSpinBox()
            self.z_scale_spinbox.setMinimum(0.01)
            self.z_scale_spinbox.setMaximum(20)
            self.z_scale_spinbox.setSingleStep(1.)
            self.z_scale_spinbox.setValue(1.)
            # self.z_scale_spinbox.valueChanged.connect(self) TODO
            z_scale_layout.addWidget(self.z_scale_spinbox)
            params_layout.addLayout(z_scale_layout)

            self.use_correction_checkbox = QCheckBox("Use correction")
            self.use_correction_checkbox.setChecked(True)
            params_layout.addWidget(self.use_correction_checkbox)

            params_widget.setLayout(params_layout)
            layout.addWidget(params_widget, 0, Qt.AlignTop)

            # Blurring parameters
            blurring_widget = CollapsibleWidget("Blurring", self)
            blurring_layout = QVBoxLayout()
            self.blur_image_checkbox = QCheckBox("Blur image")
            self.blur_image_checkbox.clicked.connect(self.on_blur_checked)
            blurring_layout.addWidget(self.blur_image_checkbox)
            self.blurring_type_widget = QWidget()
            blurring_type_layout = QVBoxLayout()
            blurring_type_layout.setContentsMargins(0, 0, 0, 0)
            blurring_type_layout.addWidget(QLabel("Blurring type", self))
            self.blurring_type_combobox = QComboBox()
            self.blurring_type_combobox.addItems(["Gaussian", "Curvature Anisotropic Diffusion"])
            self.blurring_type_combobox.setSizePolicy(QSizePolicy.Policy.Ignored, self.blurring_type_combobox.sizePolicy().verticalPolicy())
            self.blurring_type_combobox.currentTextChanged.connect(self.update_blurred_layer)
            blurring_type_layout.addWidget(self.blurring_type_combobox)
            self.blurring_type_widget.setLayout(blurring_type_layout)
            blurring_layout.addWidget(self.blurring_type_widget)
            self.cad_widget = QWidget()
            cad_layout = QVBoxLayout()
            cad_layout.setContentsMargins(0, 0, 0, 0)
            conductance_layout = QHBoxLayout()
            conductance_layout.addWidget(QLabel("Conductance"))
            delayed_updater = DelayedExecutor(self.update_blurred_layer)
            delayed_updater.processing.connect(lambda _: self.progress_widget.setVisible(blur_demo_btn.isChecked()))
            delayed_updater.processed.connect(lambda _: self.progress_widget.setVisible(False))
            self.blur_conductance_spinbox = QDoubleSpinBox()
            self.blur_conductance_spinbox.setMinimum(0)
            self.blur_conductance_spinbox.setMaximum(100)
            self.blur_conductance_spinbox.setValue(9.)
            self.blur_conductance_spinbox.valueChanged.connect(delayed_updater)
            conductance_layout.addWidget(self.blur_conductance_spinbox)
            cad_layout.addLayout(conductance_layout)

            iterations_layout = QHBoxLayout()
            iterations_layout.addWidget(QLabel("# iterations"))
            self.blur_n_iterations_spinbox = QSpinBox()
            self.blur_n_iterations_spinbox.setMinimum(1)
            self.blur_n_iterations_spinbox.setMaximum(20)
            self.blur_n_iterations_spinbox.setValue(10)
            self.blur_n_iterations_spinbox.valueChanged.connect(delayed_updater)
            iterations_layout.addWidget(self.blur_n_iterations_spinbox)
            cad_layout.addLayout(iterations_layout)
            self.cad_widget.setLayout(cad_layout)
            blurring_layout.addWidget(self.cad_widget)

            self.gauss_widget = QWidget()
            gauss_layout = QHBoxLayout()
            gauss_layout.setContentsMargins(0, 0, 0, 0)
            gauss_layout.addWidget(QLabel("Sigma", parent=self))
            self.blur_sigma_slider = QDoubleSlider(parent=self)
            self.blur_sigma_slider.setMinimum(0.)
            self.blur_sigma_slider.setMaximum(20.)
            self.blur_sigma_slider.setOrientation(Qt.Horizontal)
            self.blur_sigma_slider.sliderReleased.connect(self.update_blurred_layer)
            gauss_layout.addWidget(self.blur_sigma_slider)
            self.gauss_widget.setLayout(gauss_layout)
            blurring_layout.addWidget(self.gauss_widget)
            self.blurring_type_combobox.currentTextChanged.connect(self.on_blur_selection)

            blur_demo_btn = QPushButton("Try")
            blur_demo_btn.setCheckable(True)
            blur_demo_btn.clicked.connect(self.toggle_blur_demo)
            blurring_layout.addWidget(blur_demo_btn)
            blurring_widget.setLayout(blurring_layout)
            layout.addWidget(blurring_widget, 0, Qt.AlignTop)
            self.on_blur_checked(self.blur_image_checkbox.isChecked())
            self.crop_mask = None

            slice_annotation_widget = CollapsibleWidget("Slice annotation")
            slice_annotation_layout = QVBoxLayout()
            self.slice_segmentation_dropdown = QComboBox()
            self.slice_segmentation_dropdown.addItems(EstimationWorker.ANNOTATION_METHODS)
            slice_annotation_layout.addWidget(self.slice_segmentation_dropdown)
            slice_annotation_widget.setLayout(slice_annotation_layout)
            layout.addWidget(slice_annotation_widget, 0, Qt.AlignTop)

            self.call_button = QPushButton("Run")
            self.call_button.clicked.connect(lambda: self.call_button.setDisabled(True))
            self.call_button.clicked.connect(self.start_estimation)
            layout.addWidget(self.call_button)
            layout.addStretch()
            self.setLayout(layout)
            self._collapsible_group = CollapsibleWidgetGroup([params_widget, blurring_widget, slice_annotation_widget])

        def _add_image(self, layer_name, data, layer_args):
            if 'opacity' not in layer_args:
                layer_args['opacity'] = 0.3
            if 'blending' not in layer_args:
                layer_args['blending'] = "additive"
            if 'rgb' not in layer_args:
                layer_args['rgb'] = False
            self.viewer.add_layer(Image(data, name=layer_name, **layer_args))

        def mask_received(self, data, offset):
            offset = offset.round().astype(int)
            bb_slice = tuple(slice(offs, offs + extent) for offs, extent in zip(offset, data.shape))
            self.labels.layer.data[bb_slice][data > 0] = data[data > 0]
            self.labels.layer.refresh()

        def refresh_layer(self, layer_name):
            if layer_name not in self.viewer.layers:
                return
            layer = self.viewer.layers[layer_name]
            # layer.data = layer.data
            layer.refresh()

        def data_remover(self, name):
            self.viewer.layers.remove(name)

        def start_estimation(self):
            if self.image.layer not in self.viewer.layers:
                self.call_button.setDisabled(False)
                raise ValueError("Missing image layer")
            elif self.points_layer is None:
                self.call_button.setDisabled(False)
                raise ValueError("Missing points layer")
            elif self.labels.layer not in self.viewer.layers:
                self.call_button.setDisabled(False)
                raise ValueError("Missing labels layer")
            image = self.image.layer
            image_data = image.data
            points = self.points_layer
            points_data = points.data.copy()
            z_scale = self.z_scale_spinbox.value()
            alpha = self.alpha_slider.value()
            beta = self.beta_spinner.value()
            use_correction = self.use_correction_checkbox.isChecked()
            estimation_worker = EstimationWorker(self.viewer, self.minimal_contour_widget)
            estimation_worker.initialize(image_data, points_data, use_correction, beta, alpha, self.slice_segmentation_dropdown.currentText(), self.blur_image if self.blur_image_checkbox.isChecked() else None, z_scale)
            estimation_worker.image_data_received.connect(self._add_image)
            estimation_worker.layer_invalidated.connect(self.refresh_layer)
            estimation_worker.mask_data_received.connect(self.mask_received)
            estimation_worker.remove_layer.connect(self.data_remover)
            estimation_worker.all_done.connect(lambda: self.call_button.setDisabled(False))
            estimation_thread = QThread(self)
            estimation_worker.moveToThread(estimation_thread)
            estimation_worker.all_done.connect(estimation_thread.quit)
            estimation_thread.started.connect(estimation_worker.run)
            estimation_thread.finished.connect(estimation_worker.deleteLater)
            estimation_thread.finished.connect(estimation_thread.deleteLater)
            estimation_thread.start()
            # viewer.add_image(data)

        def hook_callbacks(self, estimator, stage, layer_name, layer_params, data_idx=None):

            def initializer(arr, idx):
                self.data_initializer(layer_name, data_idx, layer_params)(arr, idx)

            estimator.hook_stage_data_init_event(stage, initializer)
            estimator.hook_stage_iteration_event(stage, self.data_updater(layer_name))
            estimator.hook_stage_finished_event(stage, self.data_finalizer(layer_name))

        def drag_points_callback(self, layer, event):
            if layer.mode != "select":
                return
            yield
            if event.type == "mouse_move":
                if len(layer.selected_data) > 2 or len(layer.selected_data) == 0:
                    return
                if len(layer.selected_data) == 2:
                    selection_iter = iter(layer.selected_data)
                    idx1 = next(selection_iter)
                    idx2 = next(selection_iter)
                    if idx1 // 2 != idx2 // 2:
                        return
                else:
                    index = next(iter(layer.selected_data))
                    if index % 2 == 0:
                        idx1 = index
                        idx2 = index + 1
                    else:
                        idx1 = index - 1
                        idx2 = index
                p1 = layer.data[idx1]
                p2 = layer.data[idx2]
                bb = pts_2_bb(p1, p2, self.image.layer.data.shape)
                bb_slice = bb_2_slice(bb)
                image = self.image.layer.data[bb_slice]
                crop_mask = generate_sphere_mask(image.shape)
                self.image.layer.visible = False
                cropped_image = self.viewer.add_image(image*crop_mask, colormap=self.image.layer.colormap, translate=bb.min(0), rendering="attenuated_mip")
                self.viewer.layers.selection.active = layer
            yield
            while event.type == "mouse_move":
                yield
            self.viewer.layers.remove(cropped_image)
            self.update_blurred_layer()
            self.image.layer.visible = True

        def on_blur_selection(self, blur_method):
            self.gauss_widget.setVisible(blur_method == "Gaussian")
            self.cad_widget.setVisible(blur_method == "Curvature Anisotropic Diffusion")

        def on_blur_checked(self, is_checked):
            self.blurring_type_widget.setVisible(is_checked)
            self.on_blur_selection(self.blurring_type_combobox.currentText() if is_checked else "")

        def blur_image(self, image):
            blur_type = self.blurring_type_combobox.currentText()
            if blur_type == "Gaussian":
                return gaussian_filter(image, self.blur_sigma_slider.value())
            elif blur_type == "Curvature Anisotropic Diffusion":
                sitk_image = sitk.GetImageFromArray(image.astype(float))
                blurred = sitk.CurvatureAnisotropicDiffusion(
                    sitk_image,
                    conductanceParameter=self.blur_conductance_spinbox.value(),
                    numberOfIterations=self.blur_n_iterations_spinbox.value()
                )
                return sitk.GetArrayFromImage(blurred)
            elif blur_type != "":
                raise NotImplementedError("Unkown blur type %s" % blur_type)

        def toggle_blur_demo(self, is_checked):
            if len(self.points_layer.data) < 2:
                return
            image_layer = self.image.layer
            if image_layer is None:
                return
            if is_checked:
                bb = pts_2_bb(self.points_layer.data[0], self.points_layer.data[1], image_layer.data.shape)
                bb = np.clip(bb, 0, np.asarray(image_layer.data.shape) - 1).round().astype(int)
                bb_slice = bb_2_slice(bb)
                blurred_image = image_layer.data[bb_slice]
                self.crop_mask = generate_sphere_mask(blurred_image.shape)
                offset = bb.min(0)

                blurred = self.blur_image(blurred_image)
                blurred *= self.crop_mask
                self.blurred_layer = self.viewer.add_image(blurred, name="%s cropped" % image_layer.name, translate=np.add(image_layer.translate, offset),
                                   colormap=image_layer.colormap, contrast_limits=image_layer.contrast_limits)
            else:
                if self.blurred_layer:
                    self.viewer.layers.remove(self.blurred_layer)
                    self.blurred_layer = None
                self.crop_mask = None

        def update_blurred_layer(self, *args):
            if self.blurred_layer is None:
                return
            image_layer = self.image.layer
            if image_layer is None:
                return
            bb = pts_2_bb(self.points_layer.data[0], self.points_layer.data[1], image_layer.data.shape)
            bb = np.clip(bb, 0, np.asarray(image_layer.data.shape) - 1).round().astype(int)
            offset = bb.min(0)
            bb_slice = bb_2_slice(bb)
            blurred_image = image_layer.data[bb_slice]
            if self.crop_mask is None or self.crop_mask.shape != blurred_image.shape:
                self.crop_mask = generate_sphere_mask(blurred_image.shape)
            blurred = self.blur_image(blurred_image)
            blurred *= self.crop_mask
            self.blurred_layer.data = blurred
            self.blurred_layer.translate = np.add(image_layer.translate, offset)


    class ShortestPathsWidget(WidgetWithLayerList):
        shapes_data_received = Signal(str, "PyQt_PyObject", "PyQt_PyObject")

        def __init__(self, viewer: napari.Viewer):
            super().__init__(viewer, [("meeting_plane", Image), ("distance_map", Image), ("starting_points", Points)])
            self.shapes_data_received.connect(self.add_shapes)
            self.viewer = viewer
            self.estimator = MinArea.Estimator()
            layout = QVBoxLayout()

            layout.addWidget(QLabel("Extract contour from meeting plane"))
            plane_2_contour_btn = QPushButton("Extract")
            plane_2_contour_btn.clicked.connect(self.plane_2_contour)
            layout.addWidget(plane_2_contour_btn)
            line = QFrame()
            line.setFrameShape(QFrame.HLine)
            line.setFrameShadow(QFrame.Sunken)
            line.setFixedHeight(1)
            line.setStyleSheet("background-color: #414851;")
            layout.addWidget(line)

            run_button = QPushButton("Show shortest paths")
            run_button.clicked.connect(self.show_shortest_paths)
            layout.addWidget(run_button)
            self.setLayout(layout)
            self.update_layer_list(None)

        def add_shapes(self, layer_name, data, layer_args):
            if 'opacity' not in layer_args:
                layer_args['opacity'] = 0.3
            self.viewer.add_layer(Shapes(data, name=layer_name, shape_type="path", **layer_args))

        def plane_2_contour(self):
            if self.distance_map.layer is None or self.meeting_plane.layer is None:
                return
            meeting_plane_layer = self.meeting_plane.layer
            plane_mask = (meeting_plane_layer.data < 0).astype(np.uint8)
            contour = np.squeeze(cv2.findContours(plane_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)[0][0])
            contour = np.concatenate([np.fliplr(contour), np.ones([len(contour), 1])], axis=1)
            contour = contour @ meeting_plane_layer.rotate.T + meeting_plane_layer.translate
            contour_layer = Points(contour, ndim=self.distance_map.ndim, name="contour points", size=2)
            self.viewer.add_layer(contour_layer)
            for i in range(self.points_layer_dropdown.count()):
                item = self.points_layer_dropdown.itemText(i)
                if item == contour_layer.name:
                    self.points_layer_dropdown.setCurrentIndex(i)
                    break

        def calculate(self, points, distance_map, translate, rotate):
            paths = []
            for point in points:
                shortest_path = self.estimator.resolve_shortest_paths(point, distance_map).copy()
                paths.append(np.fliplr(shortest_path))
            colors = generate_label_colors(len(paths))
            self.shapes_data_received.emit("Shortest path", paths, {
                'translate': translate,
                'rotate': rotate,
                'edge_color': np.concatenate([np.asarray(colors) / 255, np.ones([len(colors), 1])], axis=1),
                'edge_width': 0.1
            })

        def show_shortest_paths(self):
            distance_map = self.distance_map.layer
            points = self.starting_points.layer
            if distance_map is None or points is None:
                return
            translation = distance_map.translate
            rotation = distance_map.rotate
            distance_map = distance_map.data
            points = np.fliplr(points.data - translation).astype(int)
            self.viewer.window.worker2 = Thread(target=self.calculate, args=[points, distance_map, translation, rotation])
            self.viewer.window.worker2.start()
else:
    MinimalSurfaceWidget = None




# cython: boundscheck = False
import time

cimport cython
import numpy as np
cimport numpy as np
cimport openmp
from libc.string cimport memcpy
from libcpp cimport bool
from libcpp.vector cimport vector
from cython.operator cimport preincrement as inc
from cython.parallel cimport prange
np.import_array()

GRADIENT_BASED = 0
INTENSITY_BASED = 2
CUSTOM_FEATURE = 3

cdef np.ndarray EMPTY = np.empty(0)

cdef extern from "Eikonal.cpp":
    pass

cdef extern from "Eikonal.h":
    cdef cppclass CVec2:
        CVec2() nogil
        CVec2(double, double) nogil
        double x, y
    cdef cppclass SWorkImg[T]:
        SWorkImg()
        void Set(int, int)
        T* operator[](int)
        int GetWidth()
        int GetHeight()
    cdef cppclass SControl:
        SControl() nogil
        int GetProgress() nogil
        int SetParam(int) nogil
        int SetParam(int, int) nogil
        bool DefineInputSet(const vector[CVec2]&, const vector[int]&) nogil
        void SetDataTerm(SWorkImg[double]*) nogil
        void SetDataTerm(SWorkImg[double]*, SWorkImg[double]*) nogil
        void GetDataTerm(SWorkImg[double]**) nogil
        void GetDataTerm(SWorkImg[double]**, SWorkImg[double]**) nogil
        void InitEnvironment(SWorkImg[double]&, SWorkImg[double]&, SWorkImg[double]&) nogil
        void InitEnvironmentAllMethods(SWorkImg[double]&, SWorkImg[double]&, SWorkImg[double]&) nogil
        void InitEnvironmentAllMethods(SWorkImg[double]&, SWorkImg[double]&, SWorkImg[double]&, SWorkImg[double]&, SWorkImg[double]&) nogil
        void InitEnvironmentRanders(SWorkImg[double]&, SWorkImg[double]&, SWorkImg[double]&) nogil
        void InitEnvironmentInhomog(SWorkImg[double]&, SWorkImg[double]&, SWorkImg[double]&) nogil
        void InitEnvironmentRandersGrad(SWorkImg[double]&, SWorkImg[double]&) nogil
        int SetNextStartStop() nogil
        void SetBoundaries(int, int, int, int) nogil
        void SetParAll() nogil
        void DistanceCalculator() nogil
        int GetReady() nogil
        vector[CVec2]& ResolvePath() nogil
        vector[CVec2]& GetMinPath() nogil
        void CleanAll() nogil
        void CalcImageQuantAllMethods() nogil

cdef class MinimalContourCalculator:
    cdef vector[SControl*] eikonals
    cdef int start_x, start_y, end_x, end_y
    cdef int param
    cdef int method
    cdef vector[int] method_pair
    cdef vector[int] progresses
    cdef SWorkImg[double] ered, egreen, eblue, egradx, egrady
    method_initialized = [False,]*4
    def __cinit__(self, np.ndarray[np.float_t, ndim=3] image, int n_points):
        self.set_image(image, np.empty((0, 0)), np.empty((0, 0)))
        self.eikonals.reserve(n_points)
        self.progresses.resize(n_points)
        self.param = 5
        self.method = 0
        self.method_pair.push_back(self.method)
        self.method_pair.push_back(self.method)
        cdef SControl* control
        cdef int i
        for i in range(n_points):
            control = new SControl()
            control.SetParam(self.param)
            control.SetParam(0, 0)
            self.eikonals.push_back(control)

    cpdef set_param(self, int param):
        cdef int i
        for i in range(self.eikonals.size()):
            self.eikonals[i].SetParam(param)
            self.eikonals[i].CleanAll()
            self.eikonals[i].CalcImageQuantAllMethods()

    cpdef set_method(self, int method=-1):
        if method == -1:
            method = self.method
        if method not in [GRADIENT_BASED, INTENSITY_BASED]:
            print("method should be one of GRADIENT_BASED(=0) or INTENSITY_BASED(=2)")
            return
        self.method = method
        self.method_pair[0] = self.method_pair[1] = method
        cdef int i
        if not self.method_initialized[method]:
            for i in range(self.eikonals.size()):
                if method == GRADIENT_BASED:
                    self.eikonals[i].InitEnvironmentRandersGrad(self.egradx, self.egrady)
                else:
                    self.eikonals[i].InitEnvironmentInhomog(self.ered, self.egreen, self.eblue)
            self.method_initialized[method] = True

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef set_boundaries(self, start_x, start_y, end_x, end_y):
        cdef int i
        for i in range(self.eikonals.size()):
            self.eikonals[i].SetBoundaries(start_x, start_y, end_x, end_y)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef set_image(self, np.ndarray[np.float_t, ndim=3] image, np.ndarray[np.float_t, ndim=2] gradx, np.ndarray[np.float_t, ndim=2] grady):
        if image is None:
            print("Image was None!")
            return
        if image.shape[2] != 3:
            print("image should have 3 channels")
            return
        cdef int w = image.shape[1]
        cdef int h = image.shape[0]
        if self.ered.GetWidth() != w or self.ered.GetHeight() != h:
            self.ered.Set(w, h)
            self.egreen.Set(w, h)
            self.eblue.Set(w, h)
        cdef double rgb_scale = 1. / 255.
        cdef int x, y
        cdef double* r_ptr
        cdef double* g_ptr
        cdef double* b_ptr
        for y in prange(h, nogil=True):
            r_ptr = self.ered[y]
            g_ptr = self.egreen[y]
            b_ptr = self.eblue[y]
            for x in range(w):
                r_ptr[x] = image[y, x, 0]
                g_ptr[x] = image[y, x, 1]
                b_ptr[x] = image[y, x, 2]
        cdef int i
        for i in range(len(self.method_initialized)):
            self.method_initialized[i] = False
        if gradx!=EMPTY and grady!=EMPTY:
            if self.egradx.GetWidth() != w or self.egrady.GetHeight() != h:
                self.egradx.Set(w, h)
                self.egrady.Set(w, h)
            for y in prange(h, nogil=True):
                r_ptr = self.egradx[y]
                g_ptr = self.egrady[y]
                for x in range(w):
                    r_ptr[x] = gradx[y, x]
                    g_ptr[x] = grady[y, x]
            for i in range(self.eikonals.size()):
                self.eikonals[i].CleanAll()
                self.eikonals[i].SetParAll()
            self.set_method()
        else:
            for i in range(self.eikonals.size()):
                self.eikonals[i].CleanAll()
                self.eikonals[i].SetParAll()
            self.set_method()

    # points are as [x, y]
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef run(
            self,
            np.ndarray[np.double_t, ndim=2] points,
            reverse_coordinates=False,
            close_path=True,
            return_segment_list=False
    ):
        if points.shape[1] != 2:
            print("Points should be 2D")
            return

        if points.shape[0] != self.eikonals.size():
            print("wrong number of points (%d to %d)" % (points.shape[0], self.eikonals.size()))
            return

        cdef int point_count = self.eikonals.size()
        cdef bool c_close_path = <bool>close_path
        if point_count < 2:
            print("At least two points should be provided")
            return
        cdef bool cancel = False

        cdef vector[CVec2] epoints
        epoints.reserve(point_count)
        cdef int idx = 0
        cdef int i, j
        cdef int X, Y
        if reverse_coordinates:
            X = 1
            Y = 0
        else:
            X = 0
            Y = 1
        for i in range(point_count):
            # TODO Check if point is out of bounds
            epoints.push_back(CVec2(points[i, X], points[i, Y]))
            inc(idx)
        for i in range(point_count):
            self.progresses[i] = 0
        cdef vector[CVec2] point_pair
        cdef CVec2 point1
        cdef CVec2 point2
        cdef int progress
        cdef SControl* eikonal
        cdef vector[vector[CVec2]] polys
        polys.resize(point_count)
        cdef int n_points = 0
        cdef int num_threads = min(point_count, openmp.omp_get_max_threads())
        for i in prange(point_count, nogil=True, num_threads=num_threads):
            if i == point_count -1:
                if not c_close_path:
                    continue
                point1 = epoints[point_count - 1]
                point2 = epoints[0]
            else:
                point1 = epoints[i]
                point2 = epoints[i+1]
            point_pair = vector[CVec2](2)
            point_pair[0] = point1
            point_pair[1] = point2

            eikonal = self.eikonals[i]
            eikonal.DefineInputSet(point_pair, self.method_pair)


            eikonal.SetNextStartStop()
            progress = self.progresses[i]
            while True:
                if cancel:
                    break
                eikonal.DistanceCalculator()
                progress = eikonal.GetProgress()
                if eikonal.GetReady() <= 0:
                    break

            eikonal.ResolvePath()
            polys[i] = eikonal.GetMinPath()
        cdef np.ndarray[np.double_t, ndim=2] segment
        if return_segment_list:
            out_list = []
            for i in range(polys.size()):
                poly = polys[i]
                segment = np.empty((poly.size(), 2), np.float64)
                for j in prange(<int>poly.size(), nogil=True):
                    segment[j, X] = poly[j].x
                    segment[j, Y] = poly[j].y
                out_list.append(segment)
            return out_list
        cdef vector[int] offsets
        offsets.reserve(polys.size())
        offsets[0] = 0
        for i in range(polys.size()):
            n_points += polys[i].size()
            if i < 1:
                continue
            offsets[i] = offsets[i-1]+polys[i-1].size()
        cdef np.ndarray[np.double_t, ndim=2] out = np.zeros([n_points, 2], dtype=np.float64)
        cdef double[:, :] out_view = out
        cdef CVec2 p
        cdef int offset
        cdef int poly_size
        for i in prange(<int>polys.size(), nogil=True, num_threads=num_threads):
            poly = polys[i]
            poly_size = poly.size()
            offset = offsets[i]
            for j in range(poly_size):
                p = poly[poly_size - 1 - j]
                out_view[offset+j, X] = p.x
                out_view[offset+j, Y] = p.y
        return out



from superqt.sliders._labeled import QLabeledSlider
from superqt.sliders._sliders import QDoubleSlider
from qtpy.QtCore import Signal, QLocale


class QLabeledDoubleSlider(QLabeledSlider):
    _slider_class = QDoubleSlider
    _slider: QDoubleSlider
    _fvalueChanged = Signal(float)
    _fsliderMoved = Signal(float)
    _frangeChanged = Signal(float, float)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.setDecimals(2)
        locale = QLocale(QLocale.Language.C)
        self._label.setLocale(locale)

    def _setValue(self, value: float):
        """Convert the value from float to int before setting the slider value."""
        self._slider.setValue(value)

    def _rename_signals(self):
        self.valueChanged = self._fvalueChanged
        self.sliderMoved = self._fsliderMoved
        self.rangeChanged = self._frangeChanged

    def decimals(self) -> int:
        return self._label.decimals()

    def setDecimals(self, prec: int):
        self._label.setDecimals(prec)




import itertools

import napari
from qtpy.QtCore import QObject, QEvent, Qt
from qtpy.QtWidgets import QWidget, QVBoxLayout, QCheckBox, QTabWidget, QPushButton, QSizePolicy
from napari import Viewer
from napari.layers import Labels, Layer, Image
from napari.qt.threading import create_worker
import numpy as np
from skimage.draw import draw
from scipy.ndimage import binary_fill_holes, find_objects
import math
from packaging import version

if version.parse(napari.__version__) >= version.parse("0.4.15"):
    try:
        from napari_bbox import BoundingBoxLayer
    except ImportError:
        BoundingBoxLayer = None
else:
    BoundingBoxLayer = None

from .interpolation_widget import InterpolationWidget
from .minimal_contour_widget import MinimalContourWidget
from .minimal_surface_widget import MinimalSurfaceWidget
from ._utils.callbacks import (
    extend_mask,
    reduce_mask,
    increase_brush_size,
    decrease_brush_size,
    scroll_to_next,
    scroll_to_prev,
    increment_selected_label,
    decrement_selected_label,
    lock_layer,
    LOCK_CHAR
)
from ._utils.help_dialog import HelpDialog
def check_connectivity(event):
    layer = event.source
    current_slice = layer._slice.image.raw


class AnnotatorWidget(QWidget):
    def __init__(self, viewer: Viewer):
        super().__init__()
        self.history_idx = 0
        layout = QVBoxLayout()
        self.fill_objects_checkbox = QCheckBox("autofill objects")
        self.fill_objects_checkbox.setChecked(True)
        self.fill_objects_checkbox.setToolTip("When drawing labels,"
                                              " close the drawn curve and fill its area after releasing the mouse")
        self._active_labels_layer = None
        self._active_bbox_layer = None
        self.drawn_region_history = dict()
        self.drawn_slice_history = dict()
        self.values_history = dict()
        self.viewer = viewer
        self.viewer.layers.selection.events.connect(self.on_layer_selection_change)
        self.viewer.layers.selection.events.connect(lock_layer)
        self.viewer.layers.events.connect(self.move_bbox_to_top)
        self.fill_objects_checkbox.clicked.connect(self.set_fill_objects)
        layout.addWidget(self.fill_objects_checkbox)

        tabs_widget = QTabWidget()
        self.interpolation_widget = InterpolationWidget(viewer)
        tabs_widget.addTab(self.interpolation_widget, "Interpolation")

        self.minimal_contour_widget = MinimalContourWidget(viewer)
        tabs_widget.addTab(self.minimal_contour_widget, "Minimal Contour")

        if MinimalSurfaceWidget is not None:
            self.minimal_surface_widget = MinimalSurfaceWidget(viewer, self.minimal_contour_widget)
            tabs_widget.addTab(self.minimal_surface_widget, "Minimal Surface")
        tabs_widget.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)
        layout.addWidget(tabs_widget)

        help_layout = QVBoxLayout()
        help_layout.setAlignment(Qt.AlignRight)
        help_button = QPushButton("?")
        help_button.setToolTip("Help")
        help_button.clicked.connect(self.show_help_window)
        help_button.setFixedSize(20, 20)
        help_layout.addWidget(help_button)
        layout.addLayout(help_layout)
        self.setLayout(layout)
        self.installEventFilter(self)
        self.set_fill_objects(True)
        self.on_layer_selection_change()
        self.init_bbox_layer()

    def eventFilter(self, source: QObject, event: QEvent) -> bool:
        if event.type() == QEvent.Hide:
            self.disconnect_all()
            return True
        elif event.type() == QEvent.Show:
            self.connect_all()
            return True
        return super().eventFilter(source, event)

    @property
    def fill_objects(self):
        return self.fill_objects_checkbox.isChecked()

    @property
    def active_labels_layer(self):
        return self._active_labels_layer

    @active_labels_layer.setter
    def active_labels_layer(self, labels_layer):
        if not isinstance(labels_layer, Labels) and labels_layer is not None:
            return
        self._active_labels_layer = labels_layer
        self.set_fill_objects(self.fill_objects)
        self.set_labels_callbacks()
        self.init_bbox_layer()

    def unset_labels_callbacks(self):
        if self.active_labels_layer is None:
            return
        if check_connectivity in self.active_labels_layer.events.data.callbacks:
            self.active_labels_layer.events.data.disconnect(check_connectivity)
        self.active_labels_layer.bind_key("Control-+", overwrite=True)(None)
        self.active_labels_layer.bind_key("Control--", overwrite=True)(None)
        self.active_labels_layer.bind_key("Q", overwrite=True)(None)
        self.active_labels_layer.bind_key("E", overwrite=True)(None)
        self.active_labels_layer.bind_key("A", overwrite=True)(None)
        self.active_labels_layer.bind_key("D", overwrite=True)(None)

    def set_labels_callbacks(self):
        if self.active_labels_layer is None:
            return
        if check_connectivity not in self.active_labels_layer.events.data.callbacks:
            self.active_labels_layer.events.data.connect(check_connectivity)
        if self.start_update_bbox not in self.active_labels_layer.events.selected_label.callbacks:
            self.active_labels_layer.events.selected_label.connect(self.start_update_bbox)
            self.active_labels_layer.events.data.connect(self.start_update_bbox)
        self.active_labels_layer.bind_key("Control-Z", overwrite=True)(self.undo)
        self.active_labels_layer.bind_key("Control-+", overwrite=True)(extend_mask)
        self.active_labels_layer.bind_key("Control--", overwrite=True)(reduce_mask)
        self.active_labels_layer.bind_key("Q", overwrite=True)(decrement_selected_label)
        self.active_labels_layer.bind_key("E", overwrite=True)(increment_selected_label)
        self.active_labels_layer.bind_key("A", overwrite=True)(scroll_to_prev(self.viewer))
        self.active_labels_layer.bind_key("D", overwrite=True)(scroll_to_next(self.viewer))
        self.active_labels_layer.bind_key("W", overwrite=True)(increase_brush_size)
        self.active_labels_layer.bind_key("S", overwrite=True)(decrease_brush_size)

    def move_bbox_to_top(self, e=None):
        if self._active_bbox_layer not in self.viewer.layers:
            return
        if e is not None\
                and e.type in ["highlight", "mode", "set_data", "data", "thumbnail", "loaded", "reordered", "editable", "translate"]:
            return
        layer_list = self.viewer.layers
        with (
            layer_list.events.moved.blocker(),
            layer_list.events.moving.blocker()
        ):
            try:
                for i in reversed(range(len(layer_list))):
                    layer = layer_list[i]
                    if layer == self._active_bbox_layer:
                        break
                    if type(layer) not in [Labels, Image]:
                        continue
                    bbox_index = layer_list.index(self._active_bbox_layer)
                    if i == bbox_index+1:
                        layer_list.move(i, bbox_index)
                    else:
                        layer_list.move(bbox_index, i)
                    break
            except KeyError:
                ...

    def start_update_bbox(self, event):
        if self._active_bbox_layer is None:
            return
        worker = create_worker(self.update_bbox, event.source)
        worker.finished.connect(self._active_bbox_layer.refresh)
        worker.start()

    def update_bbox(self, layer):
        if self._active_bbox_layer is None:
            return
        if len(self._active_bbox_layer.data) > 0:
            self._active_bbox_layer.data = []
        label = layer.selected_label
        data = layer.data
        mask = data == label
        bb_corners = find_objects(mask, max_label=1)[0]
        if bb_corners is None:
            return
        min_ = [slice_.start for slice_ in bb_corners]
        max_ = [slice_.stop - 1 for slice_ in bb_corners]
        bb = np.asarray(np.where(list(itertools.product((False, True), repeat=layer.ndim)), max_, min_))
        self._active_bbox_layer.add(bb)
        self._active_bbox_layer.properties["label"][0] = label
        self._active_bbox_layer.refresh_text()

    def set_fill_objects(self, state):
        if self.active_labels_layer is None:
            return
        if state and self.isVisible():
            if self.fill_holes not in self.active_labels_layer.mouse_drag_callbacks:
                self.active_labels_layer.mouse_drag_callbacks.append(self.fill_holes)
                self.active_labels_layer.mouse_drag_callbacks.append(self.incr_history_idx)
        elif not state:
            if self.fill_holes in self.active_labels_layer.mouse_drag_callbacks:
                self.active_labels_layer.mouse_drag_callbacks.remove(self.fill_holes)
                self.active_labels_layer.mouse_drag_callbacks.remove(self.incr_history_idx)

    def disconnect_all(self):
        if self.active_labels_layer is not None and self.fill_holes in self.active_labels_layer.mouse_drag_callbacks:
            self.active_labels_layer.mouse_drag_callbacks.remove(self.fill_holes)

    def connect_all(self):
        if self.fill_objects and self.active_labels_layer is not None and self.fill_holes not in self.active_labels_layer.mouse_drag_callbacks:
            self.active_labels_layer.mouse_drag_callbacks.append(self.fill_holes)

    def on_layer_selection_change(self, event=None):
        if event is None or event.type == "changed":
            active_layer = self.viewer.layers.selection.active
            if isinstance(active_layer, Labels):
                self.active_labels_layer = active_layer
            else:
                self.active_labels_layer = None

    @staticmethod
    def draw_line(x1, y1, x2, y2, brush_size, output):
        line_x, line_y = draw.line(x1, y1, x2, y2)
        for x, y in zip(line_x, line_y):
            cx, cy = draw.disk((x, y), math.ceil(brush_size/2+0.1))
            cx = np.clip(cx, 0, output.shape[0] - 1)
            cy = np.clip(cy, 0, output.shape[1] - 1)
            output[cx, cy] = True

    def fill_holes(self, layer: Layer, event):
        if layer.mode != "paint" or layer._ndisplay != 2:
            return
        coordinates = layer.world_to_data(event.position)
        coordinates = tuple(max(0, min(layer.data.shape[i] - 1, int(round(coord)))) for i, coord in enumerate(coordinates))
        image_coords = tuple(coordinates[i] for i in layer._dims_displayed)
        slice_dims = tuple(coordinates[i] if i not in layer._dims_displayed else slice(None) for i in range(layer.ndim))
        current_draw = np.zeros_like(layer.data[slice_dims], bool)
        start_x, start_y = prev_x, prev_y = image_coords
        cx, cy = draw.disk((start_x, start_y), layer.brush_size/2)
        cx = np.clip(cx, 0, current_draw.shape[0] - 1)
        cy = np.clip(cy, 0, current_draw.shape[1] - 1)
        current_draw[cx, cy] = True
        yield
        while event.type == 'mouse_move':
            coordinates = layer.world_to_data(event.position)
            coordinates = tuple(max(0, min(layer.data.shape[i] - 1, int(round(coord)))) for i, coord in enumerate(coordinates))
            image_coords = tuple(coordinates[i] for i in layer._dims_displayed)
            AnnotatorWidget.draw_line(prev_x, prev_y, image_coords[-2], image_coords[-1], layer.brush_size, current_draw)
            prev_x, prev_y = image_coords
            yield
        # s = np.asarray([[0, 1, 0],
        #                 [1, 1, 1],
        #                 [0, 1, 0]])
        s = None
        coordinates = layer.world_to_data(event.position)
        coordinates = tuple(
            max(0, min(layer.data.shape[i] - 1, int(round(coord)))) for i, coord in enumerate(coordinates))
        image_coords = tuple(coordinates[i] for i in layer._dims_displayed)
        prev_x, prev_y = image_coords
        AnnotatorWidget.draw_line(prev_x, prev_y, start_x, start_y, layer.brush_size, current_draw)
        cx, cy = draw.disk((prev_x, prev_y), layer.brush_size/2)
        cx = np.clip(cx, 0, current_draw.shape[0] - 1)
        cy = np.clip(cy, 0, current_draw.shape[1] - 1)
        current_draw[cx, cy] = True
        binary_fill_holes(current_draw, output=current_draw, structure=s)
        if layer.preserve_labels:
            current_draw = current_draw & (layer.data[slice_dims] == 0)
        self.drawn_region_history[self.history_idx] = current_draw
        self.drawn_slice_history[self.history_idx] = slice_dims
        self.values_history[self.history_idx] = layer.data[slice_dims][current_draw]
        layer.data[slice_dims][current_draw] = layer.selected_label
        layer.events.data()
        layer.refresh()

    def undo(self, layer):
        last_drawn_region = self.drawn_region_history.get(self.history_idx, None)
        last_drawn_slice = self.drawn_slice_history.get(self.history_idx, None)
        prev_values = self.values_history.get(self.history_idx, None)
        if self.fill_objects_checkbox.isChecked() and last_drawn_region is not None:
            layer.data[last_drawn_slice][last_drawn_region] = prev_values
            del self.drawn_region_history[self.history_idx]
            del self.drawn_slice_history[self.history_idx]
            del self.values_history[self.history_idx]
        self.decr_history_idx()
        layer.undo()
        layer.events.data()

    def incr_history_idx(self, *args):
        self.history_idx += 1

    def decr_history_idx(self, *args):
        if self.history_idx >= 0:
            self.history_idx -= 1

    def show_help_window(self):
        dialog = HelpDialog(self)
        dialog.show()

    def init_bbox_layer(self):
        if self.active_labels_layer is None or BoundingBoxLayer is None:
            return
        if self._active_bbox_layer in self.viewer.layers:
            if (self.active_labels_layer is not None and self._active_bbox_layer.ndim == self.active_labels_layer.ndim
                    or self.active_labels_layer is None and self._active_bbox_layer.ndim == 2):
                return
            with self.viewer.layers.selection.events.blocker(self.on_layer_selection_change):
                self.viewer.layers.remove(self._active_bbox_layer)
        self._active_bbox_layer = BoundingBoxLayer(name="%s active object" % LOCK_CHAR, face_color="transparent",
                                                  edge_color="green", ndim=self.active_labels_layer.ndim if self.active_labels_layer else 2)
        self._active_bbox_layer.opacity = 1.
        self._active_bbox_layer.current_properties |= {"label": 0}
        self._active_bbox_layer.text = {
            "text": "{label:d}",
            "size": 10,
            "color": "green"
        }
        with self.viewer.layers.selection.events.blocker(self.on_layer_selection_change):
            prev_selection = self.viewer.layers.selection.copy()
            self.viewer.add_layer(self._active_bbox_layer)
            if len(prev_selection) == 1 and self._active_bbox_layer not in prev_selection:
                self.viewer.layers.selection = prev_selection



import warnings
import napari
import skimage.draw
from napari.layers.labels._labels_utils import get_dtype
from napari.utils._dtype import get_dtype_limits
from skimage.morphology import binary_dilation, binary_erosion
from napari.utils.events import Event
from qtpy.QtWidgets import (
    QWidget,
    QSpinBox,
    QVBoxLayout,
    QHBoxLayout,
    QCheckBox,
    QLabel,
    QComboBox,
    QPushButton,
    QSizePolicy
)
from qtpy.QtCore import QMutex, QThread, QObject, Signal, Qt, QEvent
from qtpy.QtGui import QCursor, QKeyEvent, QPixmap, QImage
from superqt import QLargeIntSpinBox
# from napari._qt.widgets._slider_compat import QDoubleSlider
from ._utils import QDoubleSlider, ProgressWidget, WidgetWithLayerList, CollapsibleWidget, CollapsibleWidgetGroup
from ._utils.changeable_color_box import QtChangeableColorBox
from ._utils.callbacks import (
    extend_mask,
    reduce_mask,
    increase_brush_size,
    decrease_brush_size,
    scroll_to_next,
    scroll_to_prev,
    increment_selected_label,
    decrement_selected_label,
    LOCK_CHAR
)
from .image_processing_widget import ImageProcessingWidget

from ..minimal_contour import MinimalContourCalculator, FeatureManager
import numpy as np
from napari.layers import Image, Labels
from napari.layers.points._points_constants import Mode
from skimage.filters import gaussian
import scipy.fft
import colorsys

GRADIENT_BASED = 0
INTENSITY_BASED = 2

DEMO_SIZE = 200


def bbox_around_points(pts):
    p1 = pts.min(0)
    p2 = pts.max(0)
    size = p2 - p1
    from_i = p1 - size[0]*0.1 - 10
    to_i = p2 + size[0]*0.1 + 10
    return from_i, to_i


class MinimalContourWidget(WidgetWithLayerList):
    def __init__(self, viewer: napari.Viewer):
        super().__init__(viewer, [("image", Image), ("labels", Labels)])
        self.viewer = viewer
        self.calculator = MinimalContourCalculator(None, 3)
        self.progress_dialog = ProgressWidget(message="Drawing mask...")
        self.move_mutex = QMutex()
        self.draw_worker = self.DrawWorker()
        self.draw_thread = QThread()
        self.draw_worker.moveToThread(self.draw_thread)
        self.draw_worker.done.connect(self.draw_thread.quit)
        self.draw_worker.done.connect(self.set_mask)
        self.draw_worker.done.connect(lambda: self.progress_dialog.setVisible(False))
        self.draw_thread.started.connect(self.draw_worker.run)
        self._img = None
        self._prev_img_layer = self.image.layer
        self._orig_image = self.image.layer.data if self.image.layer is not None else None
        self.point_triangle = np.zeros((3, 2), dtype=np.float64) - 1  # start point, current position, end point
        self.remove_last_anchor = False
        self.last_added_with_shift = None
        self.last_segment_length = None
        self.prev_n_anchor_points = 0
        self.image.combobox.currentIndexChanged.connect(self.store_orig_image)
        self.image.combobox.currentIndexChanged.connect(self.apply_blurring)
        self.image.combobox.currentIndexChanged.connect(self.set_image)
        self.image.combobox.currentIndexChanged.connect(self.update_demo_image)
        self.prev_labels_layer = self.labels.layer
        self.feature_inverted = False
        self.test_script = False
        self.prev_timer_id = None
        self.feature_manager = FeatureManager(viewer)

        # Ctrl+Z handling
        self.change_idx = dict()
        self.prev_vals = dict()

        layout = QVBoxLayout()
        features_widget = CollapsibleWidget("Image features")
        features_layout = QVBoxLayout()
        features_layout.addWidget(QLabel("Used feature"))
        self.feature_dropdown = QComboBox()
        self.feature_dropdown.addItems(["High gradient", "High intensity", "Low intensity", "Custom"])
        self.feature_dropdown.currentIndexChanged.connect(self.on_feature_change)
        features_layout.addWidget(self.feature_dropdown)

        self.feature_editor = ImageProcessingWidget(self._img, viewer)
        self.feature_editor.setVisible(self.feature_dropdown.currentText() == "Custom")
        self.feature_editor.script_worker.done.connect(self.set_features)
        def set_test_script():
            self.test_script = True
        self.feature_editor.try_script.connect(set_test_script)
        features_layout.addWidget(self.feature_editor)

        features_layout.addWidget(QLabel("Param"))
        self.param_spinbox = QSpinBox()
        self.param_spinbox.setMinimum(1)
        self.param_spinbox.setMaximum(50)
        self.param_spinbox.setValue(5)
        self.param_spinbox.valueChanged.connect(self.on_param_change)
        self.param_spinbox.setToolTip("Increasing this parameter will cause contours to be more aligned\n"
                                      "with the selected image feature (with a lower number the contour will be\n"
                                      "closer to the Euclidean shortest path")
        self.calculator.set_param(self.param_spinbox.value())
        features_layout.addWidget(self.param_spinbox)
        features_widget.setLayout(features_layout)
        layout.addWidget(features_widget, 0, Qt.AlignTop)

        blur_widget = CollapsibleWidget("Blurring")
        blur_layout = QVBoxLayout()
        self.blur_image_checkbox = QCheckBox("Blur image")
        self.blur_image_checkbox.setChecked(True)
        self.blur_image_checkbox.clicked.connect(self.set_use_smoothing)
        self.blur_image_checkbox.setToolTip("When checked, the image will be blurred\n"
                                              "before calculating minimal contour")
        blur_layout.addWidget(self.blur_image_checkbox)

        blur_image_widget = QWidget()
        blur_image_layout = QVBoxLayout()
        blur_image_layout.setContentsMargins(0, 0, 0, 0)
        blur_image_layout.addWidget(QLabel("Blur sigma"))
        self.blur_image_slider = QDoubleSlider(parent=self)
        self.blur_image_slider.setMaximum(20)
        self.blur_image_slider.setOrientation(Qt.Horizontal)
        self.blur_image_slider.valueChanged.connect(self.update_demo_image)
        self.blur_image_slider.setVisible(self.blur_image_checkbox.isChecked())
        blur_image_layout.addWidget(self.blur_image_slider)

        demo_widget = CollapsibleWidget("Example", self)
        self.demo_image = QLabel()
        self.update_demo_image()
        demo_layout = QVBoxLayout()
        demo_layout.addWidget(self.demo_image)
        demo_widget.setLayout(demo_layout)
        blur_image_layout.addWidget(demo_widget)

        self.apply_blurring_button = QPushButton("Apply blurring")
        self.apply_blurring_button.setEnabled(self.blur_image_checkbox.isChecked())
        self.apply_blurring_button.clicked.connect(self.apply_blurring)
        blur_image_layout.addWidget(self.apply_blurring_button)
        blur_image_widget.setLayout(blur_image_layout)
        self.blur_image_widget = blur_image_widget
        blur_layout.addWidget(blur_image_widget)
        blur_widget.setLayout(blur_layout)
        layout.addWidget(blur_widget, 0, Qt.AlignTop)

        contour_widget = CollapsibleWidget("Contour")
        contour_layout = QVBoxLayout()
        smooth_contour_layout = QHBoxLayout()
        self.smooth_contour_checkbox = QCheckBox("Smooth contour")
        self.smooth_contour_checkbox.clicked.connect(lambda checked: self.smooth_contour_spinbox.setVisible(checked))
        self.smooth_contour_checkbox.setToolTip("When checked, the finished contour will be smoothed\n"
                                                "to remove minor noise using Fourier transformation.")
        self.smooth_contour_spinbox = QDoubleSlider(Qt.Horizontal)
        self.smooth_contour_spinbox.setMinimum(0.)
        self.smooth_contour_spinbox.setMaximum(1.)
        self.smooth_contour_spinbox.setValue(1.)
        # self.smooth_contour_spinbox.setTickInterval(0.05)
        self.smooth_contour_spinbox.setToolTip("Number of Fourier coefficients to approximate the contour.\n"
                                               "Lower number -> smoother contour\n"
                                               "Higher number -> more faithful to the original")
        self.smooth_contour_spinbox.setVisible(self.smooth_contour_checkbox.isChecked())
        smooth_contour_layout.addWidget(self.smooth_contour_checkbox)
        smooth_contour_layout.addWidget(self.smooth_contour_spinbox)
        contour_layout.addLayout(smooth_contour_layout)

        point_size_layout = QHBoxLayout()
        point_size_layout.addWidget(QLabel("Point size"))
        self.point_size_spinbox = QSpinBox()
        self.point_size_spinbox.setMinimum(1)
        self.point_size_spinbox.setMaximum(50)
        self.point_size_spinbox.setValue(2)
        self.point_size_spinbox.setToolTip("Point size for contour display.")
        point_size_layout.addWidget(self.point_size_spinbox)
        contour_layout.addLayout(point_size_layout)
        contour_widget.setLayout(contour_layout)
        layout.addWidget(contour_widget, 0, Qt.AlignTop)

        label_widget = CollapsibleWidget("Label options")
        label_layout = QVBoxLayout()

        self.selectionSpinBox = QLargeIntSpinBox()
        if self.labels.layer is not None:
            dtype_lims = get_dtype_limits(get_dtype(self.labels.layer))
        else:
            dtype_lims = 0, np.iinfo(int).max
        self.selectionSpinBox.setRange(*dtype_lims)
        self.selectionSpinBox.setKeyboardTracking(False)
        self.selectionSpinBox.valueChanged.connect(self.change_selected_label)
        self.selectionSpinBox.setAlignment(Qt.AlignCenter)
        self.selectionSpinBox.setToolTip("Selected label for the current 'labels' layer.")
        self.selected_id_label = QLabel()
        self.selected_id_label.setFixedSize(20, 20)
        self.selected_id_label.setAlignment(Qt.AlignCenter)
        self.selected_id_label.setSizePolicy(QSizePolicy.Fixed, self.selected_id_label.sizePolicy().verticalPolicy())

        self.selected_id_label.setWindowFlags(Qt.Window
                                              | Qt.FramelessWindowHint
                                              | Qt.WindowStaysOnTopHint)
        self.selected_id_label.setAttribute(Qt.WA_ShowWithoutActivating)
        self.selected_id_label.installEventFilter(self)
        color_layout = QHBoxLayout()
        self.colorBox = QtChangeableColorBox(self.labels.layer)
        color_layout.addWidget(self.colorBox)
        color_layout.addWidget(self.selectionSpinBox)
        self.labels.combobox.currentIndexChanged.connect(self.on_label_change)
        self.on_label_change()
        self._on_selected_label_change()
        label_layout.addLayout(color_layout)

        self.autoincrease_label_id_checkbox = QCheckBox("Increment label id")
        self.autoincrease_label_id_checkbox.setChecked(True)
        self.autoincrease_label_id_checkbox.setToolTip("When checked, selected label will be incremented\n"
                                                       "after completing the contour")
        label_layout.addWidget(self.autoincrease_label_id_checkbox)

        label_manipulation_layout = QHBoxLayout()

        reduce_mask_button = QPushButton("Reduce label")
        reduce_mask_button.clicked.connect(lambda: reduce_mask(self.labels.layer))
        reduce_mask_button.setToolTip("Reduce the selected label by one pixel around the borders\n"
                                      "using erosion")
        label_manipulation_layout.addWidget(reduce_mask_button)

        extend_mask_button = QPushButton("Extend label")
        extend_mask_button.clicked.connect(lambda: extend_mask(self.labels.layer))
        extend_mask_button.setToolTip("Extend the selected label by one pixel around the borders\n"
                                      "using dilation")
        label_manipulation_layout.addWidget(extend_mask_button)
        label_layout.addLayout(label_manipulation_layout)
        label_widget.setLayout(label_layout)
        layout.addWidget(label_widget, 0, Qt.AlignTop)
        self._collapsible_group = CollapsibleWidgetGroup([features_widget, blur_widget, contour_widget, label_widget])

        layout.addStretch()
        self.setLayout(layout)

        def change_layer_callback(num):
            def change_layer(_):
                if num-1 < len(viewer.layers):
                    viewer.layers.selection.select_only(viewer.layers[num-1])
                else:
                    warnings.warn("%d is out of range (number of layers: %d)" % (num, len(viewer.layers)))
            return change_layer
        for i in range(1, 10):
            viewer.bind_key("Control-%d" % i, overwrite=True)(change_layer_callback(i))
        self.from_e_points_layer = viewer.add_points(
            ndim=2,
            name="%s from E" % LOCK_CHAR,
            size=self.point_size_spinbox.value(),
            face_color="red",
            edge_color="red"
        )
        self.from_e_points_layer.editable = False
        self.to_s_points_layer = viewer.add_points(
            ndim=2,
            name="%s to S" % LOCK_CHAR,
            size=self.point_size_spinbox.value(),
            face_color="gray",
            edge_color="gray"
        )
        self.to_s_points_layer.editable = False
        self.output = viewer.add_points(
            ndim=2,
            name="%s temp" % LOCK_CHAR,
            size=self.point_size_spinbox.value(),
        )
        self.output.editable = False
        self.anchor_points = self.viewer.add_points(ndim=2, name="Anchors [DO NOT ALTER]", symbol="x")
        self.anchor_points.mode = "add"
        self.modifiers = None
        self.point_size_spinbox.valueChanged.connect(self.change_point_size)
        self.change_point_size(self.point_size_spinbox.value())
        viewer.dims.events.current_step.connect(self.delayed_set_image)
        viewer.dims.events.ndisplay.connect(self.set_image)
        viewer.dims.events.current_step.connect(self.update_demo_image)
        self.viewer.layers.events.connect(self.move_temp_to_top)
        self.viewer.bind_key("Control-Tab", overwrite=True)(self.swap_selection)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.viewer.window.qt_viewer.window().installEventFilter(self)
            self.viewer.window.qt_viewer.window().setMouseTracking(True)
        self.set_image()
        self.set_callbacks()
        self.move_temp_to_top()
        self.update_label_tooltip()

    def set_use_smoothing(self, use_smoothing):
        self.blur_image_widget.setVisible(use_smoothing)
        self.apply_blurring()
        self.set_image()

    def set_features(self, data):
        if self.test_script:
            self.test_script = False
            return
        if data is None:
            return
        self.calculator.set_image(data, np.empty((0, 0,)), np.empty((0, 0,)))

    def change_point_size(self, size):
        self.to_s_points_layer.size = size
        self.to_s_points_layer.selected_data = {}
        self.to_s_points_layer.current_size = size

        self.from_e_points_layer.size = size
        self.from_e_points_layer.selected_data = {}
        self.from_e_points_layer.current_size = size

        self.output.size = size
        self.output.selected_data = {}
        self.output.current_size = size
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.viewer.window.qt_viewer.canvas.native.setFocus()

    def shift_pressed(self, _):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if self.labels.layer is not None and len(self.anchor_points.data) == 0 and self.viewer.window.qt_viewer.canvas.native.hasFocus():
                self.selected_id_label.show()
        self.from_e_points_layer.face_color = "gray"
        self.from_e_points_layer.current_face_color = "gray"
        self.to_s_points_layer.face_color = "red"
        self.to_s_points_layer.current_face_color = "red"
        self.from_e_points_layer.edge_color = "gray"
        self.from_e_points_layer.current_edge_color = "gray"
        self.to_s_points_layer.edge_color = "red"
        self.to_s_points_layer.current_edge_color = "red"
        yield
        self.selected_id_label.hide()
        self.from_e_points_layer.face_color = "red"
        self.from_e_points_layer.current_face_color = "red"
        self.to_s_points_layer.face_color = "gray"
        self.to_s_points_layer.current_face_color = "gray"
        self.from_e_points_layer.edge_color = "red"
        self.from_e_points_layer.current_edge_color = "red"
        self.to_s_points_layer.edge_color = "gray"
        self.to_s_points_layer.current_edge_color = "gray"

    @property
    def ctrl_down(self):
        return self.modifiers & Qt.ControlModifier if self.modifiers is not None else False

    @property
    def shift_down(self):
        return self.modifiers & Qt.ShiftModifier if self.modifiers is not None else False

    @property
    def alt_down(self):
        return self.modifiers & Qt.AltModifier if self.modifiers is not None else False

    def ctrl_shift_pressed(self, _):
        shift_callback = self.shift_pressed(_)
        yield from shift_callback
        yield from shift_callback

    def ctrl_z_callback(self, _):
        if self.labels.layer in self.change_idx:
            idx = self.change_idx[self.labels.layer]
            vals = self.prev_vals[self.labels.layer]
            self.labels.layer.data[idx] = vals
            self.labels.layer.events.data()
            self.labels.layer.refresh()
            del self.change_idx[self.labels.layer]
            del self.prev_vals[self.labels.layer]
        else:
            warnings.warn("There's nothing to revert.")

    def esc_callback(self, _):
        self.clear_all()

    def clear_all(self):
        self.output.data = np.empty((0, 2), dtype=self.output.data.dtype)
        with self.anchor_points.events.data.blocker():
            self.anchor_points.data = np.empty((0, 2), dtype=self.output.data.dtype)
        self.prev_n_anchor_points = 0
        self.from_e_points_layer.data = np.empty((0, 2), dtype=self.output.data.dtype)
        self.to_s_points_layer.data = np.empty((0, 2), dtype=self.output.data.dtype)
        self.point_triangle[:] = -1
        self.prev_n_anchor_points = 0

    @property
    def image_data(self):
        return self._img

    @property
    def features(self):
        return self.feature_editor.features

    def on_label_change(self, _=None):
        if self.labels.layer is not None:
            dtype_lims = get_dtype_limits(get_dtype(self.labels.layer))
            self.selectionSpinBox.setValue(self.labels.layer.selected_label)
        else:
            dtype_lims = 0, np.iinfo(int).max
        self.selectionSpinBox.setRange(*dtype_lims)
        self.colorBox.layer = self.labels.layer
        if self.prev_labels_layer is not None:
            self.prev_labels_layer.events.selected_label.disconnect(self._on_selected_label_change)
            if self.on_mouse_wheel in self.prev_labels_layer.mouse_wheel_callbacks:
                self.prev_labels_layer.mouse_wheel_callbacks.remove(self.on_mouse_wheel)
            self.prev_labels_layer.bind_key("Shift", overwrite=True)(None)
            self.prev_labels_layer.bind_key("Alt", overwrite=True)(None)
            self.prev_labels_layer.bind_key("Control-+", overwrite=True)(None)
            self.prev_labels_layer.bind_key("Control--", overwrite=True)(None)
        if self.labels.layer is not None:
            self.labels.layer.events.selected_label.connect(
                self._on_selected_label_change
            )
            if self.on_mouse_wheel not in self.labels.layer.mouse_wheel_callbacks:
                self.labels.layer.mouse_wheel_callbacks.append(self.on_mouse_wheel)
            self.labels.layer.bind_key("Shift", overwrite=True)(self.shift_pressed)
            self._on_selected_label_change()
        self.prev_labels_layer = self.labels.layer

    def _on_selected_label_change(self, *args):
        """Receive layer model label selection change event and update spinbox."""
        if self.labels.layer is None:
            return
        with self.labels.layer.events.selected_label.blocker():
            value = self.labels.layer.selected_label
            self.selectionSpinBox.setValue(value)
            self.update_label_tooltip()

    def change_selected_label(self, value):
        self.labels.layer.selected_label = value
        self.selectionSpinBox.clearFocus()
        self.setFocus()

    def move_temp_to_top(self, e=None):
        if e is not None and e.type in ["highlight", "mode", "set_data", "data", "thumbnail", "loaded", "editable", "translate"]:
            return
        layer_list = self.viewer.layers
        try:
            with layer_list.events.moved.blocker(), layer_list.events.moving.blocker():
                temp_idx = layer_list.index(self.output)
                if temp_idx != len(layer_list) - 1:
                    layer_list.move(temp_idx, -1)
                if self.anchor_points is not None:
                    points_idx = layer_list.index(self.anchor_points)
                    if points_idx != len(layer_list) - 2:
                        layer_list.move(points_idx, -2)
                if self.to_s_points_layer in layer_list:
                    to_s_idx = layer_list.index(self.to_s_points_layer)
                    if to_s_idx != len(layer_list) - 3:
                        layer_list.move(to_s_idx, -3)
                if self.from_e_points_layer in layer_list:
                    from_e_idx = layer_list.index(self.from_e_points_layer)
                    if from_e_idx != len(layer_list) - 4:
                        layer_list.move(from_e_idx, -4)
        except KeyError:
            ...

    def estimate(self, image: np.ndarray):
        from_i, to_i = bbox_around_points(self.point_triangle)
        from_i = np.clip(from_i, 0, np.asarray(image.shape[:2])).astype(int)
        to_i = np.clip(to_i, 0, np.asarray(image.shape[:2])).astype(int)
        self.calculator.set_boundaries(from_i[1], from_i[0], to_i[1], to_i[0])
        results = self.calculator.run(
            self.point_triangle,
            True,
            True,
            True
        )
        return results

    def on_feature_change(self, _):
        current_text = self.feature_dropdown.currentText()
        self.feature_editor.setVisible(current_text == "Custom")
        if current_text.startswith("Low") != self.feature_inverted:
            self.feature_inverted = current_text.startswith("Low")
        self.set_image()
        self.calculator.set_method(GRADIENT_BASED if "gradient" in current_text else INTENSITY_BASED)

    def on_mouse_move(self, layer, event):
        if not self.move_mutex.tryLock():
            return
        try:
            if self.image.layer is None or self.anchor_points is None:
                return
            if layer.mode != Mode.ADD:
                return
            self.point_triangle[1] = list(self.anchor_points.world_to_data([event.position[i] for i in range(len(event.position)) if i in event.dims_displayed]))
            if np.any(self.point_triangle[1] < 0) or np.any(self.point_triangle[1] >= self.image_data.shape[:2]):
                self.point_triangle[1] = np.clip(self.point_triangle[1], 0, np.subtract(self.image_data.shape[:2], 1))
            if np.any(self.point_triangle < 0) or np.any(self.point_triangle >= self.image_data.shape[:2]):
                return
            if not self.ctrl_down:
                if self.feature_editor.isVisible():
                    if self.features is None:
                        warnings.warn("Feature image not calculated. Run script using the 'Set' button.")
                        return
                    results = self.estimate(self.features)
                else:
                    if self.image_data is None:
                        warnings.warn("Image was not set.")
                        return
                    results = self.estimate(self.image_data)
            else:
                results = [
                    np.vstack(skimage.draw.line(
                        int(self.point_triangle[1, 0]),
                        int(self.point_triangle[1, 1]),
                        int(self.point_triangle[0, 0]),
                        int(self.point_triangle[0, 1])

                    )).T,
                    np.vstack(skimage.draw.line(
                        int(self.point_triangle[2, 0]),
                        int(self.point_triangle[2, 1]),
                        int(self.point_triangle[1, 0]),
                        int(self.point_triangle[1, 1])
                    )).T
                ]
            self.from_e_points_layer.data = np.flipud(results[0])
            self.from_e_points_layer.selected_data = {}
            self.to_s_points_layer.data = np.flipud(results[1])
            self.to_s_points_layer.selected_data = {}
        finally:
            self.move_mutex.unlock()

    def on_mouse_wheel(self, _, event):
        if self.labels.layer is None:
            return
        delta = (np.sign(event.delta)).astype(int)
        if self.shift_down:
            self.labels.layer.selected_label = max(0, self.labels.layer.selected_label + delta[1])
        elif self.alt_down:
            diff = delta[0]
            diff *= min(max(1, self.labels.layer.brush_size//10), 5)
            self.labels.layer.brush_size = max(0, self.labels.layer.brush_size + diff)

    def set_callbacks(self):
        if self.anchor_points is None:
            return
        self.anchor_points.events.data.connect(self.data_event)
        self.anchor_points.events.mode.connect(self.add_mode_only)
        self.anchor_points.bind_key("Shift", overwrite=True)(self.shift_pressed)
        self.anchor_points.bind_key("Control-Shift", overwrite=True)(self.ctrl_shift_pressed)
        self.anchor_points.bind_key("Escape", overwrite=True)(self.esc_callback)
        self.anchor_points.bind_key("Control-Z", overwrite=True)(self.ctrl_z_callback)
        self.anchor_points.bind_key("Control-+", overwrite=True)(lambda _: extend_mask(self.labels.layer))
        self.anchor_points.bind_key("Control--", overwrite=True)(lambda _: reduce_mask(self.labels.layer))
        self.anchor_points.bind_key("Q", overwrite=True)(lambda _: decrement_selected_label(self.labels.layer))
        self.anchor_points.bind_key("E", overwrite=True)(lambda _: increment_selected_label(self.labels.layer))
        self.anchor_points.bind_key("A", overwrite=True)(scroll_to_prev(self.viewer))
        self.anchor_points.bind_key("D", overwrite=True)(scroll_to_next(self.viewer))
        self.anchor_points.bind_key("W", overwrite=True)(lambda _: increase_brush_size(self.labels.layer))
        self.anchor_points.bind_key("S", overwrite=True)(lambda _: decrease_brush_size(self.labels.layer))
        if self.on_double_click not in self.anchor_points.mouse_double_click_callbacks:
            self.anchor_points.mouse_double_click_callbacks.append(self.on_double_click)
        if self.on_mouse_move not in self.anchor_points.mouse_move_callbacks:
            self.anchor_points.mouse_move_callbacks.append(self.on_mouse_move)
        if self.on_right_click not in self.anchor_points.mouse_drag_callbacks:
            self.anchor_points.mouse_drag_callbacks.insert(0, self.on_right_click)
        if self.on_mouse_wheel not in self.anchor_points.mouse_wheel_callbacks:
            self.anchor_points.mouse_wheel_callbacks.append(self.on_mouse_wheel)

    def add_mode_only(self, event):
        if event.mode not in ["add", "pan_zoom"]:
            warnings.warn("Cannot change mode to %s: only 'add' and 'pan_zoom' mode is allowed" % event.mode)
            event.source.mode = "add"

    def delayed_set_image(self, *args):
        if self.prev_timer_id is not None:
            self.killTimer(self.prev_timer_id)
        self.prev_timer_id = self.startTimer(1000)

    def timerEvent(self, *args, **kwargs):
        self.apply_blurring()
        self.set_image()
        self.killTimer(self.prev_timer_id)
        self.prev_timer_id = None

    def store_orig_image(self):
        image_layer: Image = self.image.layer
        if image_layer != self._prev_img_layer:
            if self._prev_img_layer is not None:
                self._prev_img_layer.data = self._orig_image
            self._orig_image = image_layer.data.copy() if image_layer is not None else None
        self._prev_img_layer = image_layer

    def set_image(self, *args):
        image_layer: Image = self.image.layer
        if image_layer is None:
            self._img = None
            self.feature_editor.image = None
            self.feature_editor.features = None
            return
        if self.viewer.dims.ndisplay == 3:
            return
        if not image_layer.visible:
            image_layer.set_view_slice()
        self.autoincrease_label_id_checkbox.setChecked(image_layer.ndim == 2)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            image = image_layer._data_view.astype(float)
        if image.ndim == 2:
            image = np.concatenate([image[..., np.newaxis]] * 3, axis=2)
        image = (image - image.min()) / (image.max() - image.min())
        if self.feature_inverted:
            image = 1 - image
        if self.feature_editor.isVisible():
            self.feature_editor.image = image
        else:
            grad_x, grad_y = self.feature_manager.get_features(self.image.layer)
            if self.blur_image_checkbox.isChecked():
                grad_x = self.blur_image(grad_x).astype(float)
                grad_y = self.blur_image(grad_y).astype(float)
            self.calculator.set_image(image, grad_x, grad_y)
        self._img = image
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            dims_displayed = list(image_layer._dims_displayed)
        self.anchor_points.translate = image_layer.translate[dims_displayed]
        self.from_e_points_layer.translate = image_layer.translate[dims_displayed]
        self.to_s_points_layer.translate = image_layer.translate[dims_displayed]
        self.output.translate = image_layer.translate[dims_displayed]

    def data_event(self, event):
        if event.source != self.anchor_points:
            return
        if self.remove_last_anchor:
            with self.anchor_points.events.data.blocker():
                self.anchor_points.data = self.anchor_points.data[:-1]
                self.remove_last_anchor = False
                return
        anchor_data = self.anchor_points.data
        if len(anchor_data) < self.prev_n_anchor_points:
            with self.anchor_points.events.data.blocker():
                self.clear_all()
            warnings.warn("Cannot delete a single point. Cleared all anchor points")
            return
        self.prev_n_anchor_points = len(anchor_data)
        if len(anchor_data) == 0:
            return
        anchor_data[-1] = np.clip(anchor_data[-1], 0, np.subtract(self.image_data.shape[:2], 1))
        if len(anchor_data) > 1 and np.all(np.round(anchor_data[-1]) == np.round(anchor_data[-2])):
            with self.anchor_points.events.data.blocker():
                self.anchor_points.data = self.anchor_points.data[:-1]
            self.anchor_points.refresh()
            return
        self.anchor_points.refresh()
        if self.shift_down:
            self.last_added_with_shift = True
            with self.anchor_points.events.data.blocker():
                self.anchor_points.data = np.roll(self.anchor_points.data, 1, 0)
            if len(self.to_s_points_layer.data):
                self.output.data = np.concatenate([self.to_s_points_layer.data, self.output.data], 0)
                self.last_segment_length = len(self.to_s_points_layer.data)
        else:
            if len(self.from_e_points_layer.data):
                self.output.data = np.concatenate([self.output.data, self.from_e_points_layer.data], 0)
                self.last_added_with_shift = False
                self.last_segment_length = len(self.from_e_points_layer.data)

        self.point_triangle[-1] = self.anchor_points.data[0]
        self.point_triangle[0] = self.anchor_points.data[-1]

    def on_double_click(self, *args):
        if self.shift_down and len(self.from_e_points_layer.data):
            self.output.data = np.concatenate([self.output.data, self.from_e_points_layer.data], 0)
        elif not self.shift_down and len(self.to_s_points_layer.data):
            self.output.data = np.concatenate([self.to_s_points_layer.data, self.output.data], 0)
        if self.smooth_contour_checkbox.isChecked():
            self.output.data = self.smooth_fourier(self.output.data)
        self.points_to_mask()

    def on_right_click(self, layer, event: Event):
        if event.button == 2 and layer.mode == Mode.ADD:
            self.remove_last_anchor = True
            if self.last_segment_length is not None:
                with self.anchor_points.events.data.blocker():
                    if self.last_added_with_shift:
                        self.anchor_points.data = self.anchor_points.data[1:]
                        self.output.data = self.output.data[self.last_segment_length:]
                    else:
                        self.anchor_points.data = self.anchor_points.data[:-1]
                        self.output.data = self.output.data[:-self.last_segment_length]
                self.point_triangle[-1] = self.anchor_points.data[0]
                self.point_triangle[0] = self.anchor_points.data[-1]
                self.last_segment_length = None

    class DrawWorker(QObject):
        done = Signal("PyQt_PyObject")
        contour: np.ndarray
        mask_shape: tuple

        def run(self):
            mask = skimage.draw.polygon2mask(self.mask_shape, self.contour)
            mask = skimage.filters.rank.median(mask.astype(np.uint8), np.asarray([[0, 1, 0], [1, 1, 1], [0, 1, 0]])).astype(bool)
            self.done.emit(mask)

    def set_mask(self, mask):

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            idx = np.nonzero(mask)
            self.prev_vals[self.labels.layer] = self.labels.layer._slice.image.raw[mask]
            change_idx = list(self.viewer.dims.current_step)
            for i in range(2):
                change_idx[self.viewer.dims.displayed[i]] = idx[i]

            self.change_idx[self.labels.layer] = tuple(change_idx)
            self.labels.layer._slice.image.raw[mask] = self.labels.layer.selected_label
        self.labels.layer.events.data()
        self.labels.layer.refresh()
        self.clear_all()
        if self.labels.layer and self.autoincrease_label_id_checkbox.isChecked():
            self.labels.layer.selected_label += 1

    def points_to_mask(self):
        if self.image_data is None or len(self.output.data) == 0:
            return
        if self.labels.layer is None:
            warnings.warn("Missing output labels layer.")
            return
        if self.labels.layer.ndim != self.image.layer.ndim:
            warnings.warn("Shape of labels and image does not match.")
            return
        if not self.labels.layer.visible:
            self.labels.layer.set_view_slice()
        self.progress_dialog.setVisible(True)
        self.draw_worker.contour = np.asarray([np.asarray(self.labels.layer.world_to_data(self.output.data_to_world(p)))[list(self.labels.layer._dims_displayed)] for p in self.output.data])
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.draw_worker.mask_shape = self.labels.layer._data_view.shape
        self.draw_thread.start()

    def smooth_fourier(self, points):
        coefficients=max(3, round(self.smooth_contour_spinbox.value()*len(points)))
        center = points.mean(0)
        points = points - center
        tformed = scipy.fft.rfft(points, axis=0)
        tformed[0] = 0
        return scipy.fft.irfft(tformed[:coefficients], len(points), axis=0) + center

    def extend_mask(self, _):
        if self.labels.layer is None:
            return
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            labels = self.labels.layer._slice.image.raw
        mask = labels == self.labels.layer.selected_label
        mask = binary_dilation(mask)
        labels[mask] = self.labels.layer.selected_label
        self.labels.layer.events.data()
        self.labels.layer.refresh()

    def reduce_mask(self, _):
        if self.labels.layer is None:
            return
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            labels = self.labels.layer._slice.image.raw
        mask = labels == self.labels.layer.selected_label
        eroded_mask = binary_erosion(mask)
        labels[mask & ~eroded_mask] = 0
        self.labels.layer.events.data()
        self.labels.layer.refresh()

    def eventFilter(self, src: 'QObject', event: 'QEvent') -> bool:
        try:
            self.modifiers = event.modifiers()
        except:
            pass
        if src == self.selected_id_label and event.type() == QEvent.Enter:
            pos = QCursor.pos()
        elif event.type() == QEvent.Type.HoverMove:
            pos = src.mapToGlobal(event.pos())
        else:
            return False
        pos.setX(pos.x()+20)
        pos.setY(pos.y()+20)
        self.selected_id_label.move(pos)
        return False

    def update_label_tooltip(self):
        if self.labels.layer is None:
            return
        self.selected_id_label.setText(str(self.labels.layer.selected_label))
        bg_color = self.labels.layer._selected_color
        if bg_color is None:
            bg_color = np.ones(4)
        h, s, v = colorsys.rgb_to_hsv(*bg_color[:-1])
        bg_color = (255*bg_color).astype(int)
        color = (np.zeros(3) if v > 0.7 else np.ones(3)*255).astype(int)
        self.selected_id_label.setStyleSheet("background-color: rgb(%d,%d,%d); color: rgb(%d,%d,%d)" % (tuple(bg_color[:-1]) + tuple(color)))
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.viewer.window.qt_viewer.canvas.native.setFocus()

    def apply_blurring(self):
        if self.image.layer is None:
            return
        self.image.layer.data = self._orig_image.copy()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if self.blur_image_checkbox.isChecked() and self.blur_image_slider.value() > 0.:
                slice_ = tuple(
                    slice(None) if i in self.viewer.dims.displayed else self.viewer.dims.current_step[i]
                    for i in range(self.viewer.dims.ndim)
                )
                orig_image = self._orig_image[slice_]
                if self.image.layer.rgb:
                    max_vals = orig_image.max((0, 1))
                    for i in range(3):
                        if max_vals[i] == 0:
                            continue
                        self.image.layer.data[slice_ + (i,)] = self.blur_image(orig_image[..., i]).astype(self._orig_image.dtype)
                else:
                    self.image.layer.data[slice_] = self.blur_image(self._orig_image[slice_]).astype(self._orig_image.dtype)
                self.image.layer.events.data()
                self.image.layer.refresh()

    def update_demo_image(self, *args):
        if self.image.layer is None:
            demo_shape = (DEMO_SIZE, DEMO_SIZE)
            img = np.zeros((DEMO_SIZE, DEMO_SIZE, 3), np.uint8)
        else:
            im_layer: Image = self.image.layer
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                img = im_layer._data_view.astype(float)
            if all(s > DEMO_SIZE for s in img.shape if s > 3):
                demo_shape = (DEMO_SIZE, DEMO_SIZE)
                img = img[(img.shape[0]-DEMO_SIZE)//2:(img.shape[0]+DEMO_SIZE)//2, (img.shape[0]-DEMO_SIZE)//2:(img.shape[0]+DEMO_SIZE)//2]
            else:
                demo_shape = tuple(s for s in img.shape if s > 3)
            img = self.blur_image(img)
            if not im_layer.rgb:
                max_ = im_layer.contrast_limits[1]
                min_ = im_layer.contrast_limits[0]
                img = np.clip((img - min_) / (max_ - min_), 0, 1)
                img = (im_layer.colormap.map(img.ravel()).reshape(img.shape + (4,))*255)[..., :3].copy()
            img = img.astype(np.uint8)
        image = QImage(img, demo_shape[1], demo_shape[0], demo_shape[1]*3, QImage.Format.Format_RGB888)
        pixmap = QPixmap.fromImage(image, Qt.ImageConversionFlag.ColorOnly)
        self.demo_image.setPixmap(pixmap)

    def blur_image(self, img):
        return gaussian(
            img,
            sigma=self.blur_image_slider.value(),
            preserve_range=True,
            channel_axis=-1 if img.ndim == 3 else None,
            truncate=2.
        )

    def on_param_change(self, val):
        self.calculator.set_param(val)
        self.set_image()

    def swap_selection(self, viewer: napari.Viewer):
        if self.anchor_points != viewer.layers.selection.active:
            viewer.layers.selection.select_only(self.anchor_points)
        elif self.labels.layer is not None and self.labels.layer != viewer.layers.selection.active:
            viewer.layers.selection.select_only(self.labels.layer)



# from contour import *
from qtpy.QtCore import QThread, Signal
import sys
import numpy as np
import time
import os
from .settings import Settings
from ._reconstruction import reconstruct
from ._contour import initCentroid, delta_d, calcRpsvInterpolation
from ._essentials import magnitude, Contour
from .util import loadContour

class MeanThread(QThread):

    doneSignal = Signal(object)
    clearPlotSignal = Signal()
    updateSignal = Signal(float)
    rpSignal = Signal(object)
    reconSignal = Signal(object)

    def __init__(self, contours, settings=None, weights=None):
        self.settings = settings if settings else Settings()
        self.contours = contours if isinstance(contours[0], Contour)\
            else list(Contour(c.copy(), self.settings.nPoi, self.settings.resMultiplier) for c in contours)
        QThread.__init__(self)
        self.iterations = self.settings.maxIter
        self.weights = weights
    
    def __del__(self):
        self.wait()

    def run(self):

        self.updateSignal.emit(0)

        # settings for the algorithm
        settings = self.settings
        weights = np.ones(len(self.contours)) if self.weights is None else self.weights

        for i in range(len(self.contours)):
            self.contours[i].setStartingPointToLowestY()

        # init centroid at first (take average)
        startCentroid = initCentroid(self.contours, weights)

        # translate every contour
        for i_contour in range(len(self.contours)):
            self.contours[i_contour].lookup -= startCentroid
        
        # weights for interpolation

        properCentroid = startCentroid.copy()
        deltaPrev = np.zeros((1,2))

        # calculate initial mean
        imean = np.zeros((self.contours[0].parameterization.shape[0], 2))
        for i in range(len(self.contours)):
            # TODO weighting (?)
            if not self.contours[i].isClockwise():
                # if the orientation of the polygon is cclockwise, revert it
                self.contours[i].lookup = self.contours[i].lookup[::-1]
            imean += self.contours[i].lookup[self.contours[i].parameterization,:]*weights[i]
        imean /= np.sum(weights)

        # go for the maximum number of iterations (general > maxIter in settings)
        c = 0
        for i in range(self.iterations):
            print("iteration #%d" % i)
            timestamp = time.time()
            regularMean = np.zeros_like(self.contours[0].lookup[self.contours[0].parameterization, :])
            for j in range(len(self.contours)):
                regularMean += self.contours[j].lookup[self.contours[j].parameterization, :]*weights[j]
            regularMean /= np.sum(weights)
            self.contours[1].calcParams()
            # calculate the mean in RPSV space
            q_mean = calcRpsvInterpolation(self.contours, weights)
            # here we initialize the ray lengths for the reconstruction: just take the original averages
            guessRayLengths = np.zeros(self.contours[0].lookup[self.contours[0].parameterization].shape[0])
            for i_contour in range(len(self.contours)):
                contourtmp = self.contours[i_contour].lookup[self.contours[i_contour].parameterization]
                contourlengths = magnitude(contourtmp)
                guessRayLengths += contourlengths * weights[i_contour]
            guessRayLengths /= np.sum(weights)
            guessRayLengths = magnitude(regularMean)

            # lengths of the q space mean
            qraylengths = magnitude(q_mean)
            qraylengths[qraylengths<1e-99] = 1e-99

            # unit direction vectors
            dirs = q_mean/qraylengths.reshape(qraylengths.shape[0], 1) # unit direction of the mean contour points
            timestamp = time.time()
            # do the reconstruction
            r_mean_lengths, costs = reconstruct(q_mean, guessRayLengths.copy(), settings, self.rpSignal)
            # ----------------------------

            # THE mean contour in r space
            r_mean = dirs * r_mean_lengths.reshape(r_mean_lengths.shape[0], 1)

            rsqrts = qraylengths/r_mean_lengths

            # calculate delta_d displacement
            delta = delta_d(self.contours, q_mean, rsqrts)

            # calculate the differences between the current and previous displacements
            deltaDiff = deltaPrev-delta
            deltaDiff = np.sqrt(np.sum(deltaDiff*deltaDiff))

            deltaPrev = delta.copy()

            if deltaDiff<1.:
                print("centroid converged")
                self.updateSignal.emit(100)
                break
                #refCont.lookup -= delta
                #varCont.lookup -= delta
        
            for i_contour in range(len(self.contours)):
                self.contours[i_contour].lookup -= delta
            
            properCentroid += delta[0,:]
            self.updateSignal.emit(100*(i+1)/self.iterations)

        for i_contour in range(len(self.contours)):
            self.contours[i_contour].lookup += properCentroid
        r_mean += properCentroid
        self.doneSignal.emit(r_mean)


def runCommandLine(args, settings):
    contours = []
    for argind in range(1,len(args)-1):
        contours.append(loadContour(args[argind], settings.nPoi, settings.resMultiplier))
    settings.update("export", "exportName", args[len(args)-1])
    settings.updateVariables()
    meanThread = MeanThread(contours.copy(), settings)
    meanThread.start()


if __name__ == '__main__':
    if len(sys.argv)>1:
        print("running in command line mode.")
        if len(sys.argv) >= 4:
            print("number of arguments OK. Trying to load the folllowing contour files: ")
            for i in range(1,len(sys.argv)-1):
                print(sys.argv[i])
            if not os.path.exists("settings.json"):
                print("no settings.json found. Aborting...")
            else:
                stgs = Settings("settings.json")
                runCommandLine(sys.argv, stgs)
        else:
            print("number of arguments not OK. Try calling the script like \"python meanContour.py cont1.csv cont2.csv ... cont_n.csv meanName\"")
            
    else:
        print("not enough args")



import scipy.interpolate as interp
import numpy as np
import matplotlib.pyplot as plt

x = np.linspace(0,5,num=6)
y = [0,1,4,9,16,25]
finterp = interp.interp1d(x=x,y=y,kind='quadratic')


xnew = np.linspace(0,5,num=100)
plt.plot(xnew,finterp(xnew))
plt.show()


# from cEssentials import *
from ._essentials import magnitude, innerProduct
cimport numpy as np
cimport cython
from cython.parallel cimport prange
import numpy as np
import time

# determine the initial centroid, just sum up the coordinates and take their average
def initCentroid(contours, weights = None):
    r = np.zeros((1,2))
    weights = np.ones(len(contours)) if weights is None else weights
    for i in range(len(contours)):
        lut = contours[i].lookup[contours[i].parameterization]
        r += np.sum(lut, axis=0) * weights[i]
    r /= (np.sum(weights)*lut.shape[0])
    return r

def isDifferenceSmallEnough(currVelo, prevVelo):
    if prevVelo is not None:
        scurr = magnitude(currVelo)
        sprev = magnitude(prevVelo)
        diffSum = np.sum(np.abs(sprev-scurr), axis=0)/scurr.shape[0]
        if diffSum < 0.00001:
            print("reparam converged: "+str(diffSum))
            return True
        else:
            return False
    else: return False

def isMethodStuck(prevMean, energies):
    if np.abs(prevMean-np.mean(energies))<0.01:
        return True
    else:
        return False

def reparameterizeContours(contours, Settings, plotSignal=None, debug=False):
    refCont = contours[0]
    numIterations = Settings.iterations
    maxgamma = 0.24
    # maxgamma = 0.1
    refCont.calcParams()
    refRepr = refCont.getRPSV()
    nPoints = Settings.nPoi


    #fig = plt.figure()
    #camera = Camera(fig)
    sumtime = 0
    sumsmooth = 0
    g = contours[1].parameterization
    for cIter in range(1,len(contours)):
        varCont = contours[cIter]
        #varCont.parameterization = g
        stopCriterion = False
        prevDerivatives = None
        min_resolution = varCont.resMultiplier
        energies = []
        prevEnergies = -1
        costs = []
        for i in range(numIterations):
            if stopCriterion is True:
                break

            '''if i%2==0:
                varCont = contours[1]
                refCont = contours[0]
            else:
                varCont = contours[0]
                refCont = contours[1]'''

            start = time.time()
            varCont.calcParams()
            end = time.time()
            sumtime += (end-start)

            # gradient descent equation as defined in (16):
            dgamma = 0.5*(refCont.lookup[refCont.parameterization,0]*varCont.lookup[varCont.parameterization,0]+refCont.lookup[refCont.parameterization,1]*varCont.lookup[varCont.parameterization,1]+0)*(refCont.christoffel-varCont.christoffel)
            dgamma += (refCont.derivatives[:,0]*varCont.lookup[varCont.parameterization,0]+refCont.derivatives[:,1]*varCont.lookup[varCont.parameterization,1])
            dgamma -= (refCont.lookup[refCont.parameterization,0]*varCont.derivatives[:,0]+refCont.lookup[refCont.parameterization,1]*varCont.derivatives[:,1])
            dgamma *= -1
            
            maxgammaabs = np.max(np.abs(dgamma))

            if maxgammaabs<1e-99:
                maxgammaabs = 1e-99

            if maxgammaabs>maxgamma:
                dgamma /= maxgammaabs
                dgamma *= maxgamma
            
            

            #distances = np.sum(np.abs(np.sqrt(refCont.srv)-np.sqrt(varCont.srv)))
        
            varRepr = varCont.getRPSV()
            
            qdistances = np.sum(np.sqrt(innerProduct(refRepr-varRepr, refRepr-varRepr)))
            costs.append(qdistances)

            # if the optimitzer gets stuck at a point, detect it
            if (i+1)%300 == 0:
                print("energy_rp #"+str(i)+": "+str(qdistances))
                if isMethodStuck(prevEnergies, energies):
                    print("reparam stuck")
                    stopCriterion = True
                prevEnergies = np.mean(energies)
                energies.clear()
            else:
                energies.append(qdistances)


            # update with dGamma

            idx_diff = varCont.getIdxDiff(dgamma)
            g = (varCont.parameterization+idx_diff*dgamma+0.5).astype(int)
            varCont.parameterization = g

            # when we got our parameterization, make sure that we remain in range of the lookup table
            varCont.getInRangeParameterization()

            # smoothing to ensure that the order of points does not get mixed up
            startsmooth = time.time()
            varCont.smoothParameterization()
            endsmooth = time.time()
            sumsmooth += (endsmooth-startsmooth)

            #varCont.getInRangeParameterization()

            g = varCont.parameterization

            meantest = (refCont.lookup[refCont.parameterization,:]+varCont.lookup[varCont.parameterization,:])/2

            if Settings.debug is True:
                if (i%200==0):
                    cList = []
                    cList.append(refCont.lookup[refCont.parameterization[0:nPoints:15],:])
                    cList.append(varCont.lookup[varCont.parameterization[0:nPoints:15],:])
                    cList.append(meantest[0:nPoints:15,:])
                    plotSignal.emit(cList)



            if isDifferenceSmallEnough(varCont.derivatives, prevDerivatives):
                stopCriterion = True

            #avg = (refCont.lookup[refCont.parameterization]+varCont.lookup[varCont.parameterization])/2

            prevDerivatives = varCont.derivatives.copy()

        return costs


    #print("sum time spent on calcparams: "+str(sumtime))
    #print("sum time spent on smoothing: "+str(sumsmooth))

# calculate the weighted mean of contours in the rpsv space
def calcRpsvInterpolation(contours, weights):
    interp = np.zeros(contours[0].getRPSV().shape)

    # take the average in q-space: interp
    for i in range(len(contours)):
        r = contours[i].getRPSV()
        r *= weights[i]
        interp += r
    interp /= np.sum(weights)
    return interp

def calcMean(contours):
    N = len(contours)
    mean = contours[0].copy()
    for i in range(1,len(contours)):
        mean += contours[i]
    mean /= N
    return mean

# calculate centroid displacement
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef delta_d(contours, np.ndarray[np.double_t, ndim=2] q_mean, np.ndarray[np.double_t, ndim=1] rsqrts):
    cdef double energy = 0
    cdef double denominator = 0
    cdef np.ndarray[np.double_t, ndim=2] numerator = np.zeros((1,2))
    cdef np.ndarray[np.double_t, ndim=1] targetSrv
    cdef np.ndarray[np.double_t, ndim=2] targetRepr
    cdef double denominatori
    cdef np.ndarray[np.double_t, ndim=2] numeratori
    cdef int i, j

    for i in range(len(contours)):
        targetCont = contours[i]
        
        targetSrv = targetCont.getSRV()
        targetRepr = targetCont.getRPSV()
        
        denominatori = 0
        numeratori = np.zeros((1,2))

        for j in prange(targetRepr.shape[0], nogil=True):
            targetRepr[j, 0] -= q_mean[j, 0]
            targetRepr[j, 1] -= q_mean[j, 1]
            energy += targetRepr[j,0]*targetRepr[j,0]+targetRepr[j,1]*targetRepr[j,1]
            targetRepr[j, 0] *= targetSrv[j]
            targetRepr[j, 1] *= targetSrv[j]
            numerator[0, 0] += targetRepr[j, 0]
            numerator[0, 1] += targetRepr[j, 1]
            numeratori[0, 0] += targetRepr[j, 0]
            numeratori[0, 1] += targetRepr[j, 1]
            denominatori += targetSrv[j]*targetSrv[j]
            targetSrv[j] -= rsqrts[j]
            denominator += targetSrv[j]*targetSrv[j]

        targetCont.centroid[0,0] = numeratori[0,0]/denominatori
        targetCont.centroid[0,1] = numeratori[0,1]/denominatori

    senergy = np.sqrt(energy)

    if denominator>1e-99:
        delta = numerator
        delta /= denominator

    return delta



from .widget_with_layer_list import WidgetWithLayerList
from .collapsible_widget import CollapsibleWidget, CollapsibleWidgetGroup
from .napari_slider import QLabeledDoubleSlider as QDoubleSlider
from .progress_widget import ProgressWidget

__all__ = ["WidgetWithLayerList", "CollapsibleWidget", "QDoubleSlider", "ProgressWidget", "CollapsibleWidgetGroup"]



cimport numpy as np
cpdef np.ndarray[np.double_t, ndim=2] dt(np.ndarray[np.double_t, ndim=2] points, int order)
cpdef magnitude(np.ndarray[np.double_t, ndim=2] points)
cpdef np.ndarray[np.double_t, ndim=1] innerProduct(np.ndarray[np.double_t, ndim=2] curve1, np.ndarray[np.double_t, ndim=2] curve2)


from qtpy.QtCore import QMutex, QObject, QThread, Signal
import time


class DelayedQueue:
    def __init__(self):
        self._first = None
        self._second = None

    def enqueue(self, elem):
        if self._first is None:
            self._first = elem
        else:
            self._second = elem

    def get(self):
        return self._first

    def pop(self):
        self._first = self._second
        self._second = None


class DelayedExecutor(QObject):
    processing = Signal("PyQt_PyObject")
    processed = Signal("PyQt_PyObject")

    def __init__(self, func, parent=None):
        super().__init__(parent)
        self._func = func
        self._arg_queue = DelayedQueue()
        self._mutex = QMutex()
        self._worker = self.DelayedWorker(func, self._arg_queue, self._mutex, self)
        self.worker_thread = QThread()
        self._worker.moveToThread(self.worker_thread)
        self.worker_thread.destroyed.connect(self._worker.deleteLater)
        self.worker_thread.started.connect(self._worker.run)
        self.worker_thread.start()
        if parent is not None:
            parent.destroyed.connect(self.worker_thread.deleteLater)

    def __call__(self, *args, **kwargs):
        self._mutex.lock()
        self._arg_queue.enqueue((args, kwargs))
        self._mutex.unlock()

    class DelayedWorker(QObject):
        def __init__(self, func, q: DelayedQueue, mutex: QMutex, executor):
            super().__init__()
            self._func = func
            self._queue = q
            self._mutex = mutex
            self._executor = executor

        def run(self):
            while True:
                time.sleep(0.1)
                if self._mutex.tryLock():
                    args = self._queue.get()
                    self._mutex.unlock()
                    if args is not None:
                        self._executor.processing.emit(args)
                        self._func(*(args[0]), **(args[1]))
                        self._mutex.lock()
                        self._queue.pop()
                        self._mutex.unlock()
                        self._executor.processed.emit(args)
